#!/usr/bin/env python
from __future__ import with_statement
from memoize import memoize
import codecs, re, sys, os
import subprocess

LAMBDA = u'\u03bb'

NAMED_REGISTERS = ('RAX', 'RCX', 'RDX', 'RBX', 'RSP', 'RSI', 'RDI')
NAMED_REGISTER_MAPPING = dict(('r%d' % i, reg) for i, reg in enumerate(NAMED_REGISTERS))
REGISTERS = tuple(#['RAX', 'RBX', 'RCX', 'RDX', 'RSI', 'RDI', 'RBP'] + #, 'RSP'] # RSP is stack pointer?
    ['reg%d' % i for i in range(13)])
#REAL_REGISTERS = tuple(['RAX', 'RBX', 'RCX', 'RDX', 'RSI', 'RDI', 'RBP'] + #, 'RSP'] # RSP is stack pointer?
#    ['reg%d' % i for i in range(13)])
REGISTER_COLORS = ['color="black"', 'color="white",fillcolor="black"', 'color="maroon"', 'color="green"', 'fillcolor="olive"',
                   'color="navy"', 'color="purple"', 'fillcolor="teal"', 'fillcolor="silver"', 'fillcolor="gray"', 'fillcolor="red"',
                   'fillcolor="lime"', 'fillcolor="yellow"', 'fillcolor="blue"', 'fillcolor="fuschia"', 'fillcolor="aqua"']
REGISTER_COLORS = ['fillcolor="%s"' % c for c in 'aliceblue antiquewhite aquamarine azure beige bisque blue blueviolet brown cadetblue chartreuse cyan red gold deeppink darkorange'.split(' ')]
COLOR_FOR_REGISTER = dict(zip(REGISTERS, REGISTER_COLORS))

MAX_INSTRUCTION_WINDOW = 10000

def get_lines(filename):
    with codecs.open(filename, 'r', encoding='utf8') as f:
        lines = f.read().replace('\r\n', '\n')
    return [line.strip() for line in re.findall("%s '.*?[Rr]eturn [^\r\n]*" % LAMBDA, lines, flags=re.DOTALL)[0].split('\n')]

def strip_casts(text):
    return re.sub(r'\(u?int[0-9]*_t\)\s*\(?([^\)]*)\)?', r'\1', text)

def parse_lines(lines):
    lines = list(map(strip_casts, lines))
    assert lines[0][:len(LAMBDA + ' ')] == LAMBDA + ' '
    assert lines[0][-1] == ','
    ret = {}
    ret['vars'] = lines[0][len(LAMBDA + ' '):-1]
    assert lines[-1][-1] == ')'
    ret['return'] = lines[-1][:-1].replace('return ', '').replace('Return ', '')
    ret['lines'] = []
    for line in lines[1:-1]:
        datatype, varname, arg1, op, arg2 = re.findall('^(u?int[0-9]*_t) ([^ ]*) = ([^ ]*) ([^ ]*) ([^ ]*);$', line)[0]
        ret['lines'].append({'type':datatype, 'out':varname, 'op':op, 'args':(arg1, arg2), 'source':line})
    print('Compiling %d lines in groups of %d...' % (len(ret['lines']), min(MAX_INSTRUCTION_WINDOW, len(ret['lines']))))
    ret['lines'] = tuple(ret['lines'])
    split_ret = []
    for start in range(0, len(ret['lines']), MAX_INSTRUCTION_WINDOW):
        cur_ret = dict(ret)
        cur_ret['lines'] = ret['lines'][start:][:MAX_INSTRUCTION_WINDOW]
        split_ret.append(cur_ret)
    return tuple(split_ret)

def get_var_names(input_data):
    return tuple(line['out'] for line in input_data['lines'])

def get_input_var_names(input_data):
    return tuple(i for i in data['vars'].replace('%core', '').replace(',', ' ').replace('(', ' ').replace(')', ' ').replace("'", ' ').split(' ')
                 if i != '')

def get_output_var_names(input_data):
    return tuple(i for i in data['return'].replace(',', ' ').replace('(', ' ').replace(')', ' ').split(' ')
                 if i != '')

def line_of_var(input_data, var):
    retv = [line for line in input_data['lines'] if line['out'] == var]
    if len(retv) > 0: return retv[0]
    return {'out': var, 'args':tuple(), 'op': 'INPUT', 'type':'uint64_t'}

def make_data_dependencies(input_data):
    input_var_names = get_input_var_names(input_data)
    dependencies = dict((var, tuple()) for var in input_var_names)
    for line in input_data['lines']:
        dependencies[line['out']] = tuple(arg for arg in line['args']
                                          if arg[0] not in '0123456789')
    return dependencies
def make_reverse_data_dependencies(dependencies):
    reverse_dependencies = dict((k, []) for k in dependencies.keys())
    for k, v in dependencies.items():
        for arg in v:
            reverse_dependencies[arg].append(k)
    return reverse_dependencies

def get_low_or_high(obj, low_or_high):
    assert(low_or_high in ('low', 'high'))
    if obj['op'] == 'COMBINE':
        if low_or_high == 'low': return obj['deps'][0]
        if low_or_high == 'high': return obj['deps'][1]
    else:
        return {'out':obj['out'] + '_' + low_or_high, 'style':'', 'deps':(obj,), 'op':'GET_' + low_or_high.upper(), 'type':'uint64_t', 'extra_out':tuple(o + '_' + low_or_high for o in obj['extra_out']), 'rev_deps':tuple()}

def add_combine_low_high(objs):
    for obj in objs:
        if obj['type'] == 'uint128_t':
            obj_low = get_low_or_high(obj, 'low')
            obj_high = get_low_or_high(obj, 'high')
            obj_new = {'out':obj['out'], 'style':'', 'deps':(obj_low, obj_high), 'op':'COMBINE', 'type':'uint128_t', 'extra_out':obj['extra_out'], 'rev_deps':obj['rev_deps']}
            obj['out'] += '_tmp'
            obj['rev_deps'] = (obj_low, obj_high)
            obj_high['rev_deps'] = (obj_new,)
            obj_low['rev_deps'] = (obj_new,)
            for rdep in obj_new['rev_deps']:
                rdep['deps'] = tuple(d if d is not obj else obj_new
                                     for d in rdep['deps'])


def split_graph(objs):
    for obj in objs:
        if obj['op'] == '&' and obj['type'] == 'uint64_t' and len(obj['deps']) == 1 and obj['deps'][0]['type'] == 'uint128_t' and obj['deps'][0]['op'] == 'COMBINE':
            combine_node = obj['deps'][0]
            low = combine_node['deps'][0]
            obj['deps'] = (low,)
            low['rev_deps'] = tuple(list(low['rev_deps']) + [obj])
        if obj['op'] == '+' and obj['type'] == 'uint128_t' and len(obj['rev_deps']) == 2 and obj['rev_deps'][0]['op'] == 'GET_LOW' and obj['rev_deps'][1]['op'] == 'GET_HIGH':
            for tmp in ('_tmp', '_temp'):
                if obj['out'][-len(tmp):] == tmp:
                    obj['out'] = obj['out'][:-len(tmp)]
            obj_low, obj_high = obj['rev_deps']
            obj_carry = {'out':'c' + obj['out'], 'style':'', 'deps':(obj_low,), 'op':'GET_CARRY', 'type':'CARRY', 'extra_out':tuple(), 'rev_deps':(obj_high,)}
            assert(len(obj_low['deps']) == 1)
            assert(len(obj_high['deps']) == 1)
            assert(obj_low['type'] == 'uint64_t')
            assert(obj_high['type'] == 'uint64_t')
            obj_low['deps'], obj_high['deps'] = [], [obj_carry]
            obj_low['op'] = '+'
            obj_high['op'] = '+'
            for dep in obj['deps']:
                if dep['type'] == 'uint64_t':
                    obj_low['deps'].append(dep)
                    dep['rev_deps'] = tuple(d if d is not obj else obj_low
                                            for d in dep['rev_deps'])
                elif dep['type'] == 'uint128_t':
                    dep_low, dep_high = get_low_or_high(dep, 'low'), get_low_or_high(dep, 'high')
                    obj_low['deps'].append(dep_low)
                    obj_high['deps'].append(dep_high)
                    dep_low['rev_deps'] = tuple(list(dep_low['rev_deps']) + [obj_low])
                    dep_high['rev_deps'] = tuple(list(dep_high['rev_deps']) + [obj_high])
                else:
                    assert(False)
            obj_low['deps'], obj_high['deps'] = tuple(obj_low['deps']), tuple(obj_high['deps'])
            obj_low['rev_deps'] = list(obj_low['rev_deps']) + [obj_carry]
            obj['deps'] = tuple()
            obj['rev_deps'] = tuple()

def collect_ac_buckets(graph):
    to_process = list(graph['out'].values())
    while len(to_process) > 0:
        line, to_process = to_process[0], to_process[1:]
        if line['op'] == '+':
            args = list(line['deps'])
            new_args = []
            while len(args) > 0:
                arg, args = args[0], args[1:]
                if arg['op'] == '+' and len(arg['rev_deps']) == 1 and line['type'] == 'uint128_t':
                    line['extra_out'] = tuple(sorted(list(line['extra_out']) + [arg['out']] + list(arg['extra_out'])))
                    for arg_arg in arg['deps']:
                        arg_arg['rev_deps'] = (line,)
                        args.append(arg_arg)
                else:
                    new_args.append(arg)
            line['deps'] = tuple(new_args)
        to_process += list(line['deps'])

def get_objects(start, ret=None):
    if ret is None: ret = {}
    for node in start:
        if node['out'] in ret.keys(): continue
        ret[node['out']] = node
        get_objects(node['deps'], ret=ret)
    return ret

def prune(start):
    objs = get_objects(start)
    for var in objs.keys():
        objs[var]['rev_deps'] = tuple(objs[arg] for arg in sorted(objs.keys())
                                      if any(node['out'] == var for node in objs[arg]['deps']))

def to_graph(input_data):
    objs = dict((var, {'out':var, 'style':''}) for var in list(get_input_var_names(input_data)) + list(get_var_names(input_data)))
    for var in get_input_var_names(input_data):
        objs[var]['deps'] = tuple()
        objs[var]['op'] = 'INPUT'
        objs[var]['type'] = 'uint64_t'
        objs[var]['extra_out'] = tuple()
    for line in input_data['lines']:
        var = line['out']
        objs[var]['extra_out'] = tuple()
        objs[var]['op'] = line['op']
        objs[var]['type'] = line['type']
        objs[var]['deps'] = tuple(objs[arg] for arg in line['args'] if arg in objs.keys())
    for var in objs.keys():
        objs[var]['rev_deps'] = tuple(objs[arg] for arg in sorted(objs.keys())
                                      if any(node['out'] == var for node in objs[arg]['deps']))
    graph = {'out':dict((var, objs[var]) for var in get_output_var_names(input_data)),
             'in':dict((var, objs[var]) for var in get_input_var_names(input_data)) }
    collect_ac_buckets(graph)
    add_combine_low_high(objs.values())
    split_graph(objs.values())
    prune(tuple(graph['out'].values()))
    #split_graph(objs)
    return graph


def adjust_bits(input_data, graph):
    for line in input_data['lines']:
        if line['type'] == 'uint128_t':
            graph = graph.replace(line['out'], line['out'] + '_128')
    return graph

def fill_node(node, color='red'):
    node['style'] = ', style="filled", fillcolor="%s"' % color

def fill_deps(node, color='red'):
    fill_node(node)
    for dep in node['deps']:
        fill_deps(dep, color=color)

def fill_subgraph(in_node, color='red'):
    #print((in_node['out'], in_node['op'], [d['out'] for d in in_node['rev_deps']]))
    fill_node(in_node, color=color)
    if in_node['op'] != '+':
        fill_deps(in_node, color=color)
        for rdep in in_node['rev_deps']:
            fill_subgraph(rdep, color=color)

def is_temp(node):
    for tmp in ('_tmp', '_temp'):
        if node['out'][-len(tmp):] == tmp:
            return True
    return False

def is_allocated_to_reg(full_map, node):
    return node['out'] in full_map.keys() and all(reg in REGISTERS for reg in full_map[node['out']].split(':'))

def deps_allocated(full_map, node):
    if node['op'] == 'INPUT': return True
    if node['out'] not in full_map.keys(): return False
    return all(deps_allocated(full_map, dep) for dep in node['deps'])

# returns {cur_map with new_name->reg}, still_free_temps, still_free_list, all_temps, freed, new_buckets, emit_vars
def allocate_node(existing, node, *args):
    cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars = args
    free_temps = list(free_temps)
    free_list = list(free_list)
    all_temps = list(all_temps)
    full_map = dict(existing)
    cur_map = dict(cur_map)
    freed = list(freed)
    new_buckets = list(new_buckets)
    emit_vars = list(emit_vars)
    full_map.update(cur_map)
    def do_ret():
        return cur_map, tuple(free_temps), tuple(free_list), tuple(all_temps), tuple(freed), tuple(new_buckets), tuple(emit_vars)
    def do_free(var):
        for reg in full_map[var].split(':'):
            if reg in all_temps:
                if reg not in free_temps:
                    free_temps.append(reg)
            elif reg in REGISTERS:
                if reg not in free_list:
                    print('freeing %s from %s' % (reg, var))
                    free_list.append(reg)
    def do_free_deps(node):
        full_map.update(cur_map)
        if node['out'] in full_map.keys():
            for dep in node['deps']:
                if dep['out'] in freed or dep['out'] not in full_map.keys(): continue
                if not is_allocated_to_reg(full_map, dep): continue
                if (all(deps_allocated(full_map, rdep) for rdep in dep['rev_deps']) or
                    all(reg in all_temps for reg in full_map[dep['out']].split(':'))):
                    do_free(dep['out'])
                    freed.append(dep['out'])
    if node['out'] in full_map.keys():
        do_free_deps(node)
        return do_ret()
    #print('alloc: %s (of %d)' % (node['out'], len(free_list)))
    if node['op'] in ('GET_HIGH', 'GET_LOW') and len(node['deps']) == 1 and len(node['deps'][0]['rev_deps']) <= 2 and all(n['op'] in ('GET_HIGH', 'GET_LOW') for n in node['deps'][0]['rev_deps']) and is_allocated_to_reg(full_map, node['deps'][0]):
        reg_idx = {'GET_LOW':0, 'GET_HIGH':1}[node['op']]
        cur_map[node['out']] = full_map[node['deps'][0]['out']].split(':')[reg_idx]
        emit_vars.append(node)
        return do_ret()
    if len(node['deps']) == 1 and len(node['deps'][0]['rev_deps']) == 1 and is_allocated_to_reg(full_map, node['deps'][0]) and node['type'] == node['deps'][0]['type']:
        cur_map[node['out']] = full_map[node['deps'][0]['out']]
        emit_vars.append(node)
        return do_ret()
    if len(node['deps']) == 0 and node['op'] == 'INPUT':
        assert(node['type'] == 'uint64_t')
        cur_map[node['out']] = 'm' + node['out'] # free_list.pop()
        emit_vars.append(node)
        return do_ret()
    if is_temp(node):
        num_reg = {'uint64_t':1, 'uint128_t':2}[node['type']]
        # TODO: make this more efficient by allowing re-use of
        # dependnecies which are no longer needed (which are currently
        # only reaped after this node is assigned)
        while len(free_temps) < num_reg:
            reg = free_list.pop()
            free_temps.append(reg)
            all_temps.append(reg)
        cur_map[node['out']] = ':'.join(free_temps[:num_reg])
        free_temps = free_temps[num_reg:]
        emit_vars.append(node)
        do_free_deps(node)
        return do_ret()
    if node['op'] == '+' and node['type'] == 'uint64_t' and len(node['extra_out']) > 0:
        cur_map[node['out']] = free_list.pop()
        emit_vars.append(node)
        new_buckets.append(node)
        do_free_deps(node)
        return do_ret()
    if node['op'] == '*' and node['type'] == 'uint64_t' and len(node['deps']) == 1:
        dep = node['deps'][0]
        assert(dep['out'] in full_map.keys())
        if is_allocated_to_reg(full_map, dep) and \
           all(rdep is node or (is_allocated_to_reg(full_map, rdep) and full_map[rdep['out']] != full_map[dep['out']])
               for rdep in dep['rev_deps']):
            cur_map[node['out']] = full_map[dep['out']]
            freed += [dep['out']]
        else:
            cur_map[node['out']] = free_list.pop()
        emit_vars.append(node)
        return do_ret()
    raw_input([node['out'], node['op'], node['type'], [(dep['out'], full_map.get(dep['out'])) for dep in node['deps']]])
    return do_ret()

def allocate_deps(existing, node, *args):
    for dep in node['deps']:
        args = allocate_deps(existing, dep, *args)
    return allocate_node(existing, node, *args)

def allocate_subgraph(existing, node, *args):
    if node['op'] != '+':
        args = allocate_deps(existing, node, *args)
    else:
        args = allocate_node(existing, node, *args)
    if node['op'] != '+':
        for rdep in node['rev_deps']:
            args = allocate_subgraph(existing, rdep, *args)
    return args

def annotate_with_alloc(objs, mapping):
    for obj in objs:
        if obj['out'] in mapping.keys():
            obj['reg'] = ' (' + mapping[obj['out']] + ')'
        else:
            obj['reg'] = ''

def get_plus_deps(nodes, ops=('+',), types=('uint128_t',), seen=None):
    if seen is None: seen = set()
    for node in nodes:
        for dep in node['deps']:
            if dep['out'] in seen: continue
            seen.add(dep['out'])
            if dep['op'] in ops and dep['type'] in types:
                yield dep
            for dep in get_plus_deps([dep], ops=ops, types=types, seen=seen):
                yield dep

def print_nodes(objs):
    for var in sorted(objs.keys(), key=(lambda s:(int(s.strip('cx_lowhightmp')), s))):
        yield '    %s [label="%s%s" %s];\n' % (objs[var]['out'], ' + '.join(sorted([objs[var]['out']] + list(objs[var]['extra_out']))), objs[var]['reg'], objs[var]['style'])
def print_deps(objs):
    for var in sorted(objs.keys()):
        for dep in objs[var]['deps']:
            yield '    %s -> %s [ label="%s" ] ;\n' % (dep['out'], objs[var]['out'], objs[var]['op'])

def push_allocate(existing, nodes, *args, **kwargs):
    if 'seen' not in kwargs.keys(): kwargs['seen'] = set()
    full_map = dict(existing)
    for node in nodes:
        if node['out'] in kwargs['seen']: continue
        cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars = args
        free_temps = list(free_temps)
        free_list = list(free_list)
        all_temps = list(all_temps)
        cur_map = dict(cur_map)
        freed = list(freed)
        new_buckets = list(new_buckets)
        emit_vars = list(emit_vars)
        full_map.update(cur_map)
        if node['out'] in full_map.keys() and node['op'] == '+' and all(d['out'] not in full_map.keys() for d in node['rev_deps']) and set(d['op'] for d in node['rev_deps']) == set(('&', 'COMBINE', 'GET_CARRY')):
            and_node = [d for d in node['rev_deps'] if d['op'] == '&'][0]
            carry_node = [d for d in node['rev_deps'] if d['op'] == 'GET_CARRY'][0]
            combine_node = [d for d in node['rev_deps'] if d['op'] == 'COMBINE'][0]
            high_node = [d for d in combine_node['deps'] if d is not node][0]
            assert(len(combine_node['rev_deps']) == 1)
            shr_node = combine_node['rev_deps'][0]
            assert(shr_node['op'] == '>>')
            assert(shr_node['out'] not in full_map.keys())
            assert(len(combine_node['deps']) == 2)
            assert(all(d['out'] in full_map.keys() for d in combine_node['deps']))
            cur_map[carry_node['out']] = 'c0'
            emit_vars.append(carry_node)
            cur_map[combine_node['out']] = ':'.join(full_map[d['out']] for d in combine_node['deps'])
            emit_vars.append(combine_node)
            assert(high_node['out'] in full_map.keys())
            cur_map[shr_node['out']] = full_map[high_node['out']]
            emit_vars.append(shr_node)
            cur_map[and_node['out']] = full_map[node['out']]
            emit_vars.append(and_node)
            fill_node(combine_node)
            fill_node(carry_node)
            fill_node(shr_node)
            fill_node(and_node)
            freed += [node['out'], carry_node['out'], high_node['out'], combine_node['out']]
        elif node['out'] in full_map.keys() and len(node['rev_deps']) == 1 and all(d['out'] not in full_map.keys() for d in node['rev_deps']) and len(node['rev_deps'][0]['deps']) == 1 and node['type'] == node['rev_deps'][0]['type']:
            next_node = node['rev_deps'][0]
            cur_map[next_node['out']] = full_map[node['out']]
            emit_vars.append(next_node)
            fill_node(next_node)
            full_map.update(cur_map)
            freed += [node['out']]
        elif node['out'] not in full_map.keys() and len(node['rev_deps']) == 2 and len(node['deps']) == 2 and all(d['out'] not in full_map.keys() for d in node['rev_deps']) and all(d['out'] in full_map.keys() for d in node['deps']) and node['type'] == 'uint64_t' and all(d['type'] == 'uint64_t' for d in node['rev_deps']) and all(d['type'] == 'uint64_t' for d in node['deps']):
            from1, from2 = node['deps']
            to1, to2 = node['rev_deps']
            assert(full_map[from1['out']] != full_map[from2['out']])
            cur_map[node['out']] = full_map[from1['out']]
            emit_vars.append(node)
            cur_map[to1['out']] = full_map[from1['out']]
            emit_vars.append(to1)
            cur_map[to2['out']] = full_map[from2['out']]
            emit_vars.append(to2)
            fill_node(node)
            fill_node(to1)
            fill_node(to2)
            full_map.update(cur_map)
            freed += [node['out'], from1['out'], from2['out']]
        elif node['out'] not in full_map.keys() and len(node['rev_deps']) == 0 and len(node['deps']) == 2 and all(d['out'] not in full_map.keys() for d in node['rev_deps']) and all(d['out'] in full_map.keys() for d in node['deps']) and node['type'] == 'uint64_t' and all(d['type'] == 'uint64_t' for d in node['rev_deps']) and all(d['type'] == 'uint64_t' for d in node['deps']):
            from1, from2 = node['deps']
            assert(full_map[from1['out']] != full_map[from2['out']])
            cur_map[node['out']] = full_map[from1['out']]
            emit_vars.append(node)
            fill_node(node)
            full_map.update(cur_map)
            freed += [from1['out'], from2['out']]
        full_map.update(cur_map)
        args = (cur_map, tuple(free_temps), tuple(free_list), tuple(all_temps), tuple(freed), tuple(new_buckets), tuple(emit_vars))
        kwargs['seen'].add(node['out'])
        args = push_allocate(existing, node['rev_deps'], *args, **kwargs)
    return args

def allocate_one_subtree(in_nodes, possible_nodes, existing, *args):
    cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars = args
    existing, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars \
        = dict(existing), dict(cur_map), list(free_temps), list(free_list), list(all_temps), tuple(freed), tuple(new_buckets), tuple(emit_vars)
    args = (cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars)
    sorted_nodes = []
    for node in possible_nodes:
        try:
            lens = [len([rd for rd in d['rev_deps'] if rd['out'] not in existing.keys()]) for d in node['deps']]
            temp_cur_map, temp_free_temps, temp_free_list, temp_all_temps, temp_freed, temp_new_buckets, temp_emit_vars = allocate_subgraph(existing, node, *args)
            if set(temp_free_temps) != set(temp_all_temps):
                print(('BAD', node['out'], temp_cur_map, temp_free_temps, temp_free_list, temp_all_temps, temp_freed))
            sorted_nodes.append(((len(temp_free_list),
                                  -min(lens),
                                  -max(lens),
                                  -len(temp_new_buckets),
                                  len(temp_free_temps),
                                  -int(node['out'].strip('x_lowhightemp'))),
                                 node))
        except IndexError:
            print('Too many reg: %s' % node['out'])
    sorted_nodes = tuple(reversed(sorted(sorted_nodes, key=(lambda v: v[0]))))
    print([(n[0], n[1]['out']) for n in sorted_nodes])
    node = sorted_nodes[0][1]
    possible_nodes = [n for n in possible_nodes if n is not node]
    print('Allocating for %s' % node['out'])
    args = allocate_subgraph(existing, node, *args)
    fill_subgraph(node)
    args = push_allocate(existing, in_nodes, *args)
    cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars = args
    return possible_nodes, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars


def print_graph(graph, allocs):
    objs = get_objects(graph['out'].values())
    annotate_with_alloc(objs.values(), allocs)
    body = ''.join(print_nodes(objs))
    body += ''.join(print_deps(objs))
    body += ''.join('    in -> %s ;\n' % node['out'] for node in graph['in'].values())
    body += ''.join('    %s -> out ;\n' % node['out'] for node in graph['out'].values())
    return ('digraph G {\n' + body + '}\n')

def fix_emit_vars(emit_vars):
    ret = []
    waiting = []
    seen = set()
    for node in emit_vars:
        waiting.append(node)
        new_waiting = []
        for wnode in waiting:
            if wnode['out'] in seen:
                continue
            elif all(dep['out'] in seen for dep in wnode['deps']):
                ret.append(wnode)
                seen.add(wnode['out'])
            else:
                new_waiting.append(wnode)
        waiting = new_waiting
    while len(waiting) > 0:
        print('Waiting on...')
        print(list(sorted(node['out'] for node in waiting)))
        new_waiting = []
        for wnode in waiting:
            if wnode['out'] in seen:
                continue
            elif all(dep['out'] in seen for dep in wnode['deps']):
                ret.append(wnode)
                seen.add(wnode['out'])
            else:
                new_waiting.append(wnode)
        waiting = new_waiting
    return tuple(ret)

def print_input(reg_out, mem_in):
    #return '%s <- LOAD %s;\n' % (reg_out, mem_in)
    #return '"mov %%[%s], %%[%s]\\n\\t"\n' % (mem_in, reg_out)
    return ""

def print_val(reg):
    if reg.upper() in NAMED_REGISTERS:
        return '%%%s' % reg
    if reg[:2] == '0x':
        return '$%s' % reg
    return '%%[%s]' % reg

def print_load_specific_reg(reg, specific_reg='rdx'):
    ret = ''
    #ret += '"mov %%%s, %%[%s_backup]\\t\\n" // XXX: How do I specify that a particular register should be %s?\n' % (specific_reg, specific_reg, specific_reg)
    if reg != specific_reg:
        ret += '"mov %s, %s\\t\\n"\n' % (print_val(reg), print_val(specific_reg))
    return ret, specific_reg
def print_unload_specific_reg(specific_reg='rdx'):
    ret = ''
    #ret += '"mov %%[%s_backup], %%%s\\t\\n" // XXX: How do I specify that a particular register should be %s?\n' % (specific_reg, specific_reg, specific_reg)
    return ret
#def get_arg_reg(d):
#    return 'arg%d' % d
def print_load(reg, can_clobber=tuple(), dont_clobber=tuple()):
    assert(not isinstance(can_clobber, str))
    assert(not isinstance(dont_clobber, str))
    can_clobber = [i for i in reversed(can_clobber) if i not in dont_clobber]
    if reg in REGISTERS:
        return ('', reg)
    else:
        cur_reg = can_clobber.pop()
        ret = '"mov %s, %s\\t\\n"\n' % (print_val(reg), print_val(cur_reg))
        return (ret, cur_reg)

def print_mulx(reg_out_low, reg_out_high, rx1, rx2, src):
    #return '%s:%s <- MULX %s, %s; // %s\n' % (reg_out_low, reg_out_high, rx1, rx2, src)
    ret = ''
    ret2, actual_rx1 = print_load_specific_reg(rx1, 'rdx')
    assert(rx2 != actual_rx1)
    ret3, actual_rx2 = print_load(rx2, can_clobber=[reg_out_high, reg_out_low], dont_clobber=[actual_rx1])
    ret += ret2 + ret3 + ('"mulx %s, %s, %s\\t\\n" // %s\n' % (print_val(actual_rx2), print_val(reg_out_high), print_val(reg_out_low), src))
    ret += print_unload_specific_reg('rdx')
    return ret

def print_mov_bucket(reg_out, reg_in, bucket):
    #return '%s <- MOV %s; // bucket: %s\n' % (reg_out, reg_in, bucket)
    #ret, reg_in = print_load(reg_in, can_clobber=[reg_out])
    return ('"mov %s, %s\\t\\n" // bucket: %s\n' % (print_val(reg_in), print_val(reg_out), bucket))

def print_mov(reg_out, reg_in):
    #return '%s <- MOV %s;\n' % (reg_out, reg_in)
    #ret, reg_in = print_load(reg_in)
    return ('"mov %s, %s\\t\\n"\n' % (print_val(reg_in), print_val(reg_out)))

def print_load_constant(reg_out, imm):
    assert(imm[:2] == '0x')
    return ('"mov $%s, %s\\t\\n"\n' % (imm, print_val(reg_out)))

LAST_CARRY = None

def print_mul_by_constant(reg_out, reg_in, constant, src):
    #return '%s <- MULX %s, %s; // %s\n' % (ret_out, reg_in, constant, src)
    ret = ''
    if constant == '0x13':
        ret += ('// FIXME: lea for %s\n' % src)
    assert(constant[:2] == '0x')
    return ret + \
        print_load_constant('rdx', constant) + \
        print_mulx(reg_out, 'rdx', 'rdx', reg_in, src)

def print_adx(reg_out, rx1, rx2, bucket):
    #return '%s <- ADX %s, %s; // bucket: %s\n' % (reg_out, rx1, rx2, bucket)
    assert(rx1 == reg_out)
    ret, rx2 = print_load(rx2, dont_clobber=[rx1])
    return ret + ('"adx %%[%s], %%[%s]\\t\\n" // bucket: %s\n' % (rx2, reg_out, bucket))

def print_add(reg_out, cf, rx1, rx2, bucket):
    #return '%s, (%s) <- ADD %s, %s; // bucket: %s\n' % (reg_out, cf, rx1, rx2, bucket)
    global LAST_CARRY
    assert(reg_out == rx1)
    #assert(LAST_CARRY is None or LAST_CARRY == cf)
    LAST_CARRY = cf
    ret, rx2 = print_load(rx2, dont_clobber=[rx1])
    return ret + ('"add %%[%s], %%[%s]\\t\\n" // bucket: %s\n' % (rx2, reg_out, bucket))

def print_adc(reg_out, cf, rx1, rx2, bucket):
    #return '%s, (%s) <- ADC (%s), %s, %s; // bucket: %s\n' % (reg_out, cf, cf, rx1, rx2, bucket)
    assert(reg_out == rx1)
    ret = ''
    global LAST_CARRY
    if LAST_CARRY != cf:
        ret += 'ERRRRRRROR: %s != %s\n' % (LAST_CARRY, cf)
        LAST_CARRY = cf
    ret2, rx2 = print_load(rx2, dont_clobber=[rx1])
    ret += ret2
    return ret + ('"adc %%[%s], %%[%s]\\t\\n" // bucket: %s\n' % (rx2, reg_out, bucket))

def print_adcx(reg_out, cf, bucket):
    #return '%s <- ADCX (%s), %s, 0x0; // bucket: %s\n' % (reg_out, cf, reg_out, bucket)
    assert(LAST_CARRY == cf)
    return ('"adcx $0, %%[%s]\\t\\n" // bucket: %s\n' % (reg_out, bucket))

def print_and(reg_out, rx1, rx2, src):
    #return '%s <- AND %s, %s; // %s\n' % (reg_out, rx1, rx2, src)
    global LAST_CARRY
    LAST_CARRY = None
    if reg_out != rx1:
        return print_mov(reg_out, rx1) + print_and(reg_out, reg_out, rx2, src)
    else:
        if rx2[:2] == '0x':
            return ('"and $%s, %%[%s]\\t\\n" // %s\n' % (rx2, reg_out, src))
        else:
            ret, rx2 = print_load(rx2, can_clobber=[reg_out], dont_clobber=[rx1])
            return ret + ('"and %%[%s], %%[%s]\\t\\n" // %s\n' % (rx2, reg_out, src))

def print_shr(reg_out, rx1, imm, src):
    #return '%s <- SHR %s, %s; // %s\n' % (reg_out, rx1, imm, src)
    global LAST_CARRY
    LAST_CARRY = None
    assert(rx1 == reg_out)
    assert(imm[:2] == '0x')
    return ('"shr $%s, %%[%s]\\t\\n" // %s\n' % (imm, reg_out, src))

def print_shrd(reg_out, rx_low, rx_high, imm, src):
    #return '%s <- SHR %s, %s; // %s\n' % (reg_out, rx1, imm, src)
    global LAST_CARRY
    LAST_CARRY = None
    if rx_low != reg_out and rx_high == reg_out:
        return print_mov('rdx', rx_low) + \
            print_mov(rx_high, rx_low) + \
            print_mov(rx_low, 'rdx') + \
            print_shrd(reg_out, rx_high, rx_low, imm, src)
    assert(rx_low == reg_out)
    assert(imm[:2] == '0x')
    return ('"shrd $%s, %%[%s], %%[%s]\\t\\n" // %s\n' % (imm, rx_low, rx_high, src))


def schedule(input_data, existing, emit_vars):
    ret = ''
    buckets_seen = set()
    buckets_carried = set()
    emit_vars = fix_emit_vars(emit_vars)
    ret += ('// Convention is low_reg:high_reg\n')
    for node in emit_vars:
        if node['op'] == 'INPUT':
            ret += print_input(existing[node['out']], node['out'])
        elif node['op'] == '*' and len(node['deps']) == 2:
            assert(len(existing[node['out']].split(':')) == 2)
            out_low, out_high = existing[node['out']].split(':')
            ret += print_mulx(out_low, out_high,
                              existing[node['deps'][0]['out']],
                              existing[node['deps'][1]['out']],
                              '%s = %s * %s'
                              % (node['out'],
                                 node['deps'][0]['out'],
                                 node['deps'][1]['out']))
        elif node['op'] == '*' and len(node['deps']) == 1:
            extra_arg = [arg for arg in line_of_var(data, node['out'])['args'] if arg[:2] == '0x'][0]
            ret += print_mul_by_constant(existing[node['out']],
                                         existing[node['deps'][0]['out']],
                                         extra_arg,
                                         '%s = %s * %s'
                                         % (node['out'],
                                            node['deps'][0]['out'],
                                            extra_arg))
        elif node['op'] == '&' and len(node['deps']) == 1:
            extra_arg = [arg for arg in line_of_var(data, node['out'])['args'] if arg[:2] == '0x'][0]
            ret += print_and(existing[node['out']],
                             existing[node['deps'][0]['out']],
                             extra_arg,
                             '%s = %s & %s'
                             % (node['out'],
                                node['deps'][0]['out'],
                                extra_arg))
        elif node['op'] == '>>' and len(node['deps']) == 1 and node['deps'][0]['op'] == 'COMBINE':
            extra_arg = [arg for arg in line_of_var(data, node['out'])['args'] if arg[:2] == '0x'][0]
            ret += print_shrd(existing[node['out']],
                              existing[node['deps'][0]['deps'][0]['out']],
                              existing[node['deps'][0]['deps'][1]['out']],
                              extra_arg,
                              '%s = %s:%s >> %s'
                              % (node['out'],
                                 node['deps'][0]['deps'][0]['out'],
                                 node['deps'][0]['deps'][1]['out'],
                                 extra_arg))
        elif node['op'] == '>>' and len(node['deps']) == 1 and node['deps'][0]['type'] == 'uint64_t':
            extra_arg = [arg for arg in line_of_var(data, node['out'])['args'] if arg[:2] == '0x'][0]
            ret += print_shr(existing[node['out']],
                             existing[node['deps'][0]['deps'][0]['out']],
                             extra_arg,
                             '%s = %s >> %s'
                             % (node['out'],
                                node['deps'][0]['deps'][0]['out'],
                                extra_arg))
        elif node['op'] in ('GET_HIGH', 'GET_LOW'):
            if node['rev_deps'][0]['out'] not in buckets_seen:
                ret += print_mov_bucket(existing[node['rev_deps'][0]['out']],
                                        existing[node['out']],
                                        ' + '.join(sorted([node['rev_deps'][0]['out']] + list(node['rev_deps'][0]['extra_out']))))
                buckets_seen.add(node['rev_deps'][0]['out'])
            elif node['op'] == 'GET_HIGH':
                ret += print_adx(existing[node['rev_deps'][0]['out']],
                                 existing[node['rev_deps'][0]['out']],
                                 existing[node['out']],
                                 ' + '.join(sorted([node['rev_deps'][0]['out']] + list(node['rev_deps'][0]['extra_out']))))
            elif node['op'] == 'GET_LOW':
                carry = 'c' + node['rev_deps'][0]['out'][:-len('_low')]
                if node['rev_deps'][0]['out'] not in buckets_carried:
                    ret += print_add(existing[node['rev_deps'][0]['out']],
                                     carry,
                                     existing[node['rev_deps'][0]['out']],
                                     existing[node['out']],
                                     ' + '.join(sorted([node['rev_deps'][0]['out']] + list(node['rev_deps'][0]['extra_out']))))
                    buckets_carried.add(node['rev_deps'][0]['out'])
                else:
                    ret += print_adc(existing[node['rev_deps'][0]['out']],
                                     carry,
                                     existing[node['rev_deps'][0]['out']],
                                     existing[node['out']],
                                     ' + '.join(sorted([node['rev_deps'][0]['out']] + list(node['rev_deps'][0]['extra_out']))))
        elif node['op'] in ('GET_CARRY',):
            carry = 'c' + node['rev_deps'][0]['out'][:-len('_high')]
            ret += print_adcx(existing[node['rev_deps'][0]['out']],
                              carry,
                              ' + '.join(sorted([node['rev_deps'][0]['out']] + list(node['rev_deps'][0]['extra_out']))))
        elif node['op'] == '+' and len(node['extra_out']) > 0:
            pass
        elif node['op'] == '+' and len(node['deps']) == 2 and node['type'] == 'uint64_t':
            ret += print_adx(existing[node['out']],
                             existing[node['deps'][0]['out']],
                             existing[node['deps'][1]['out']],
                             '%s = %s + %s'
                             % (node['out'],
                                node['deps'][0]['out'],
                                node['deps'][1]['out']))
        elif node['op'] in ('COMBINE',):
            pass
        else:
            raw_input((node['out'], node['op']))
        if node['op'] not in ('GET_HIGH', 'GET_LOW', 'COMBINE', 'GET_CARRY'):
            for rdep in node['rev_deps']:
                if len(rdep['extra_out']) > 0 and rdep['op'] == '+':
                    if rdep['out'] not in buckets_seen:
                        ret += print_mov_bucket(existing[rdep['out']],
                                                existing[node['out']],
                                                ' + '.join(sorted([rdep['out']] + list(rdep['extra_out']))))
                        buckets_seen.add(rdep['out'])
                    elif 'high' in rdep['out']:
                        ret += print_adx(existing[rdep['out']],
                                         existing[rdep['out']],
                                         existing[node['out']],
                                         ' + '.join(sorted([rdep['out']] + list(rdep['extra_out']))))
                    elif 'low' in rdep['out']:
                        carry = 'c' + rdep['out'][:-len('_low')]
                        if rdep['out'] not in buckets_carried:
                            ret += print_add(existing[rdep['out']],
                                             carry,
                                             existing[rdep['out']],
                                             existing[node['out']],
                                             ' + '.join(sorted([rdep['out']] + list(rdep['extra_out']))))
                            buckets_carried.add(rdep['out'])
                        else:
                            ret += print_adc(existing[rdep['out']],
                                             carry,
                                             existing[rdep['out']],
                                             existing[node['out']],
                                             ' + '.join(sorted([rdep['out']] + list(rdep['extra_out']))))
                    else:
                        assert(False)
    return ret

def inline_schedule(sched, input_vars, output_vars):
    KNOWN_CONSTRAINTS = dict(('r%sx' % l, l) for l in 'abcd')
    def int_or_zero_key(v):
        orig = v
        v = v.strip('abcdefghijklmnopqrstuvwxyz')
        if v.isdigit(): return (int(v), orig)
        return (0, orig)
    variables = list(sorted(set(list(re.findall('%\[([a-zA-Z0-9_]*)\]', sched)) +
                                list(re.findall('%([a-zA-Z0-9_]+)', sched))),
                            key=int_or_zero_key))
    mems, variables = [i for i in variables if i[:2] == 'mx'], [i for i in variables if i[:2] != 'mx']
    special_reg, variables = [i for i in variables if i.upper() in NAMED_REGISTERS], [i for i in variables if i.upper() not in NAMED_REGISTERS]
    transient_regs, output_regs = [i for i in variables if i not in output_vars.values()], [i for i in variables if i in output_vars.keys()]
    available_registers = ['r%d' % i for i in range(16)
                           if ('r%d' % i) not in NAMED_REGISTER_MAPPING.keys() or NAMED_REGISTER_MAPPING['r%d' % i].lower() not in special_reg]
    for reg in output_regs:
        sched = sched.replace('%%[%s]' % reg, '%%[r%s]' % output_vars[reg])
    renaming = dict((from_reg, to_reg) for from_reg, to_reg in zip(transient_regs, available_registers[-len(transient_regs):]))
    for from_reg, to_reg in renaming.items():
        sched = sched.replace('%%[%s]' % from_reg, '%%%s' % to_reg)
    transient_regs = [renaming[reg] for reg in transient_regs]
    ret = ''
    ret += 'asm (\n'
    ret += sched
    ret += ': ' + ', '.join(['[r%s] "=&r" (%s)' % (output_vars[reg], output_vars[reg]) for reg in output_regs]) + '\n'
    ret += ': ' + ', '.join(['[%s] "m" (%s)' % (reg, input_vars[reg]) for reg in input_vars]) + '\n'
    ret += ': ' + ', '.join(['"cc"'] +
                            ['"%s"' % reg for reg in special_reg] +
                            ['"%s"' % reg for reg in transient_regs]) + '\n'
    ret += ');\n'
    return ret

data_list = parse_lines(get_lines('femulDisplay.log'))
for i, data in enumerate(data_list):
    graph = to_graph(data)
    #buckets = tuple(sorted(get_plus_deps(graph['out'].values()),
    #                       key=(lambda n:len(list(get_plus_deps([n]))))))
    possible_nodes = dict((n['out'], n)
                          for in_obj in graph['in'].values()
                          for n in in_obj['rev_deps']
                          if n['op'] == '*')
    for var, node in list(possible_nodes.items()):
        possible_nodes.update(dict((n['out'], n)
                                   for n in node['rev_deps']
                                   if n['op'] == '*'))
    possible_nodes = list(sorted(possible_nodes.items()))
    possible_nodes = [n for v, n in possible_nodes]
    in_nodes = tuple(graph['in'].values())
    existing, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars = {}, {}, tuple(), tuple(REGISTERS), tuple(), tuple(), tuple(), tuple()
    objs = get_objects(graph['out'].values())
    def vars_for(var, rec=True):
        pre_ret = [n['out'] for n in objs[var]['rev_deps']]
        ret = [v for v in pre_ret if 'tmp' in v]
        if rec:
            for v in pre_ret:
                if 'tmp' not in v:
                    ret += list(vars_for(v, rec=False))
        return tuple(ret)
    def vars_for_bucket(var):
        if '_' not in var:
            return tuple(list(vars_for_bucket(var + '_low')) + list(vars_for_bucket(var + '_high')))
        ret = []
        for dep in objs[var]['deps']:
            if dep['op'] in ('GET_HIGH', 'GET_LOW'):
                assert(len(dep['deps']) == 1)
                assert('tmp' in dep['deps'][0]['out'])
                ret.append(dep['deps'][0]['out'])
        return tuple(ret)
#    for var in list(vars_for('x10')) + list(vars_for('x11')) + list(vars_for('x9')) + list(vars_for('x7')) + list(vars_for('x5')): # tuple(): #('x20_tmp', 'x49_tmp', 'x51_tmp', 'x55_tmp', 'x53_tmp'):
    for var in list(vars_for_bucket('x56')) + list(vars_for_bucket('x71')) + list(vars_for_bucket('x74')) + list(vars_for_bucket('x77')) + list(vars_for_bucket('x80')): # + list(vars_for('x11')) + list(vars_for('x9')) + list(vars_for('x7')) + list(vars_for('x5')): # tuple(): #('x20_tmp', 'x49_tmp', 'x51_tmp', 'x55_tmp', 'x53_tmp'):
        print(var)
        cur_possible_nodes = [n for n in possible_nodes if n['out'] == var]
        cur_possible_nodes, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars \
            = allocate_one_subtree(in_nodes, cur_possible_nodes, existing, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars)
        existing.update(cur_map)
        cur_map = {}
    for count in range(0 * 16):
        print(count)
        possible_nodes, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars \
            = allocate_one_subtree(in_nodes, possible_nodes, existing, cur_map, free_temps, free_list, all_temps, freed, new_buckets, emit_vars)
        existing.update(cur_map)
        cur_map = {}
    #my_node = [n for n in possible_nodes if n['out'] == 'x36_tmp'][0]
    #fill_subgraph(my_node)
    #possible_nodes, cur_map, free_temps, free_list, all_temps \
    #        = allocate_one_subtree([my_node], existing, cur_map, free_temps, free_list, all_temps)
    #mul_node = possible_nodes[0]
    #print([n['out'] for n in mul_node['deps']])
    #cur_map, free_temps, free_list, all_temps = allocate_subgraph(existing, mul_node, cur_map, free_temps, free_list, all_temps)
    sched = inline_schedule(schedule(data, existing, emit_vars),
                            dict((existing[n['out']], n['out']) for n in graph['in'].values()),
                            dict((existing[n['out']], n['out']) for n in graph['out'].values()))
    #fill_deps(buckets[0])
    deps = adjust_bits(data, print_graph(graph, existing))
    with codecs.open('femulData%d.dot' % i, 'w', encoding='utf8') as f:
        f.write(deps)
    with codecs.open('femulDisplayScheduled%d.log' % i, 'w', encoding='utf8') as f:
        f.write(sched)
    for fmt in ('png', 'svg'):
        subprocess.call(['dot', '-T%s' % fmt, 'femulData%d.dot' % i, '-o', 'femulData%d.%s' % (i, fmt)])



#!/usr/bin/env python3
import sys, re, csv

SILENCE_MISSING_REAL_USER_WARNING = True

def warn(text):
    sys.stderr.write(text + '\n')

def get_input(fname):
    if fname == '-':
        return sys.stdin.readlines()
    else:
        with open(fname, 'r') as f:
            return f.readlines()

def postformat(lines):
    i = 0
    while i < len(lines):
        if i + 1 < len(lines):
            if lines[i].startswith('Testing ') and lines[i+1].startswith('Tactic call ran for'):
                yield lines[i] + lines[i+1]
                i += 2
                continue
        if lines[i].startswith('Testing '):
            yield lines[i]
            i += 1
        elif lines[i].strip() in ('', 'No such index') or lines[i].startswith('No params'):
            i += 1
        else:
            warn('Could not recognize: %s' % lines[i])
            i += 1

def eval_numexpr(numexpr):
    # copying from https://stackoverflow.com/a/25437733/377022
    numexpr = re.sub(r"\.(?![0-9])", "", numexpr) # purge any instance of '.' not followed by a number
    return eval(numexpr, {'__builtins__':None})

REGS = {'kind': r'(UnsaturatedSolinas|WordByWordMontgomery)',
        'bitwidth': r'bitwidth = ([0-9]+)',
        'prime': r'(?:UnsaturatedSolinas|WordByWordMontgomery)[\s"]*([^" ]+)',
        'method': r'method = ([a-zA-Z0-9_-]+)',
        'descr': r'\|}\s*\)[\s"]*([^:]+)',
        'real': r'(?:ran for |real:\s*)([0-9\.]*) s',
        'user': r'(?:secs \(|user:\s*)([0-9\.]*)(?:u,| s)',
        'index': r'(?:index = ([0-9]+)|WordByWordMontgomery)',
        'nlimbs': r'(?:[ \.]n := ([0-9]+)|WordByWordMontgomery)'}
def get_data(line):
    ret = {}
    bad = False
    for key, reg in REGS.items():
        res = re.search(reg, line)
        if res is None:
            if key in ('real', 'user'):
                if SILENCE_MISSING_REAL_USER_WARNING: return None
                else: bad = True
            warn('Could not find %s (%s) in %s' % (key, reg, line))
        else:
            ret[key] = res.groups()[0]
            if ret[key] is None: ret[key] = ''
            ret[key] = ret[key].replace('"', '').strip()
    if ret['index'] != '': ret['index'] = '(%s)' % ret['index']
    ret['prime_str'] = ret['prime']
    ret['prime'] = eval_numexpr(ret['prime'].replace('^', '**'))
    if ' with ' in ret['descr']:
        ret['descr1'], ret['descr2'] = ret['descr'].split(' with ')
    else:
        ret['descr1'], ret['descr2'] = ret['descr'], ''
    del ret['descr']
    if bad: return None
    return ret

def get_data_for_graph(lines, key_is_good):
    def key_for_sort(key):
        return ('Gallina' in key, 'cbv' in key, 'vm_compute' in key, 'extraction' in key, key)
    data = {}
    for line in map(get_data, lines):
        if line is None: continue
        prime = line['prime']
        if not prime in data.keys(): data[prime] = {}
        key = ('%(kind)s x%(bitwidth)s %(index)s | %(descr1)s with %(descr2)s using %(method)s' % line).strip()
        new_data = {'real': line['real'], 'user': line['user']}
        if key in data[prime]: warn('Overwriting (%s) data[%s][%s] = %s with %s' % (line['prime_str'], prime, key, str(data[prime][key]), str(new_data)))
        data[prime][key] = new_data
    pre_all_keys = list(sorted(set(key for dat in data.values() for key in dat.keys() if key_is_good(key)), key=key_for_sort))
    all_keys = list(sorted(set(key.split(' | ')[1] for key in pre_all_keys), key=key_for_sort))
    extra_key_parts = list(sorted(set(key.split(' | ')[0] for key in pre_all_keys)))
    return data, all_keys, extra_key_parts


POSSIBLE_COLUMNS = ('perf_old_vm_times',
                    'perf_new_vm_times',
                    'perf_new_extraction_times',
                    'perf_old_cbv_times',
	            'perf_new_extraction_over_old_vm',
                    'perf_new_vm_over_old_vm',
                    'perf_old_vm_over_old_vm',
	            'perf_new_extraction_over_new_extraction',
                    'perf_new_vm_over_new_extraction',
                    'perf_old_vm_over_new_extraction')

def lines_to_rows(lines, for_graph=False, real_or_user='real', only=None, **kwargs):
    def key_is_good(key):
        if only is not None and only.replace('-', ' ').replace('_', ' ') not in key: return False
        if 'lazy' in key or '(1)' in key or 'native_compute' in key: return False
        if 'Pipeline' in key and 'NBE' not in key: return False
        if 'Pipeline' in key and 'cbv' in key: return False
        if 'Pipeline' not in key and 'GallinaAxOf' not in key: return False
        return True
    if for_graph:
        data, all_keys, extra_key_parts = get_data_for_graph(lines, key_is_good)
        yield ['prime'] + [key + (' (%s, s)' % real_or_user) for key in all_keys] + ['extra info']
        defaultv = None
        default = {'real':defaultv, 'user':defaultv}
        for prime in sorted(data.keys(), key=int):
            for extra_key in extra_key_parts:
                yield [prime] + [data[prime].get(extra_key + ' | ' + short_key, default)[real_or_user] for short_key in all_keys] + [extra_key]
    elif any(kwargs[col] for col in POSSIBLE_COLUMNS):
        data, all_keys, extra_key_parts = get_data_for_graph(lines, key_is_good)
        defaultv = None
        default = {'real':defaultv, 'user':defaultv}
        def format_prime(p): return p
        def format_time(t): return t
        def format_time_ratio(n, d): return float(n) / float(d)
        def get_div_row(p, n=None, d=None):
            if p is None or n is None or d is None or '' in (p, n, d): return None
            return [format_prime(p), format_time_ratio(n, d)]
        def get_row(p, t=None):
            if p is None or t is None or '' in (p, t): return
            return [format_prime(p), format_time(t)]
        arg_to_key_filter = {
            'perf_old_vm_times': (lambda short_key: 'Pipeline' not in short_key and 'vm_compute' in short_key and 'precomputed_decision_tree' in short_key),
            'perf_new_vm_times': (lambda short_key: 'Pipeline' in short_key and 'vm_compute' in short_key and 'precomputed_decision_tree' in short_key),
            'perf_new_extraction_times': (lambda short_key: 'Pipeline' in short_key and 'extraction' in short_key and 'precomputed_decision_tree' in short_key),
            'perf_old_cbv_times': (lambda short_key: 'Pipeline' not in short_key and 'cbv' in short_key and 'precomputed_decision_tree' in short_key)
        }
        div_arg_to_key_filters = dict(
            (key, (arg_to_key_filter['perf_' + n_key + '_times'], arg_to_key_filter['perf_' + d_key + '_times']))
            for key, (n_key, d_key) in
            [(key, key.replace('perf_', '').split('_over_')) for key in POSSIBLE_COLUMNS if '_over_' in key]
        )

        for prime in sorted(data.keys(), key=int):
            for extra_key in extra_key_parts:
                extra_args = [(short_key, data[prime].get(extra_key + ' | ' + short_key, default)[real_or_user]) for short_key in all_keys]
                for arg, key_filter in arg_to_key_filter.items():
                    if kwargs[arg]:
                        row = get_row(prime, *[v for short_key, v in extra_args if key_filter(short_key)])
                        if row is not None: yield row
                for arg, (n_key_filter, d_key_filter) in div_arg_to_key_filters.items():
                    if kwargs[arg]:
                        row = get_div_row(prime,
                                          *[v for short_key, v in extra_args if n_key_filter(short_key)],
                                          *[v for short_key, v in extra_args if d_key_filter(short_key)])
                        if row is not None: yield row
    else:
        keys = ['prime', 'user', 'real', 'kind', 'bitwidth', 'descr1', 'descr2', 'method', 'index', 'nlimbs', 'prime_str']
        yield list(keys)
        for data in map(get_data, lines):
            if data is None: continue
            yield [data[k] for k in keys]

def writecsv(outfname, lines, **kwargs):
    rows = lines_to_rows(lines, **kwargs)
    def do_write(csvfile):
        writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL, lineterminator='\n')
        writer.writerows(rows)
    if outfname == '-': do_write(sys.stdout)
    else:
        with open(outfnname, 'w', newline='') as csvfile:
            do_write(csvfile)

def writetxt(outfname, lines, **kwargs):
    rows = lines_to_rows(lines, **kwargs)
    def do_write(txtfile):
        txtfile.write('\n'.join(' '.join(map(str, row)) for row in rows))
    if outfname == '-': do_write(sys.stdout)
    else:
        with open(outfnname, 'w', newline='') as txtfile:
            do_write(txtfile)

if __name__ == '__main__':
    fnames = sys.argv[1:]
    outfname = '-'
    if '-o' in fnames:
        i = fnames.index('-o')
        outfname = fnames[i+1]
        del fnames[i+1]
        del fnames[i]
    only = None
    for i, fname in enumerate(fnames):
        if '--only-' in fname and fname.startswith('--'):
            if only is not None:
                raise Exception('Only one argument can start with -- and contain --only-')
            fnames[i], only = fname.split('--only-')
    kwargs = dict([('for_graph', False)] + [(i, False) for i in POSSIBLE_COLUMNS])
    for key in kwargs.keys():
        arg = '--' + key.replace('_', '-')
        if arg in fnames:
            kwargs[key] = True
            del fnames[fnames.index(arg)]
    kwargs['only'] = only
    txt = False
    if '--txt' in fnames:
        txt = True
        del fnames[fnames.index('--txt')]
    while '--file-list' in fnames:
        file_list_file = fnames[fnames.index('--file-list')+1]
        del fnames[fnames.index('--file-list')+1]
        del fnames[fnames.index('--file-list')]
        with open(file_list_file, 'r') as f:
            file_list = [i.strip() for i in f.readlines() if i.strip()]
        fnames.extend(file_list)
    lines = [line
             for fname in fnames
             for line in postformat(get_input(fname))]
    if txt:
        writetxt(outfname, lines, **kwargs)
    else:
        writecsv(outfname, lines, **kwargs)



#!/usr/bin/python3

curves = {}

try:
    while True:
        line = input()
        curve, variant, time = line.split()
        if curve not in curves:
            curves[curve] = {}
        curves[curve][variant] = float(time)
except EOFError:
    pass

for curve, variants in curves.items():
    def compare(fiat, other):
        if fiat in variants and other in variants:
            print(curve, variants[other] / variants[fiat])

    compare('fiat_solinas64', 'gmpvar64')
    compare('fiat_solinas32', 'gmpvar32')
    compare('fiat_montgomery64', 'gmpvar64')
    compare('fiat_montgomery32', 'gmpvar32')



#!/usr/bin/env python3
import sys, re, os

SCRIPT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))
GENERATED_DIRECTORY = os.path.join(SCRIPT_DIRECTORY, 'generated')
EXTRACTION_DIRECTORY = os.path.realpath(os.path.join(SCRIPT_DIRECTORY, '..', '..', '..', 'ExtractionOCaml'))
BITWIDTHS = ('32', '64')
MKINDS = (('unsaturated_solinas', 'UnsaturatedSolinas', (' 0%nat', ' 1%nat')), ('word_by_word_montgomery', 'WordByWordMontgomery', ('',)))
VKINDS = ('GallinaAxOf', 'GallinaAxComputedOf', 'PipelineOf', 'GallinaDefOf')

def get_input(fname):
    if fname == '-':
        return sys.stdin.readlines()
    else:
        with open(fname, 'r') as f:
            return f.readlines()

def filter_lines(lines):
    for line in lines:
        line = re.sub('#.*', '', line).strip(' \n').replace(' ', '')
        if line != '':
            yield line

def write_if_different(fname, contents):
    if os.path.exists(fname):
        with open(fname, 'r') as f:
            old_contents = f.read()
        if old_contents == contents: return
    with open(fname, 'w') as f:
        f.write(contents)

def prime_to_fname(prime, bitwidth, descr, ext=''):
    return 'p%s__x%s__%s%s' % (prime.replace('^', '').replace('-', 'm').replace('+', 'p').replace('*', 't'), bitwidth, descr.replace('%nat', '').replace(' ', '_'), ext)

def write_files(prime):
    if not os.path.exists(GENERATED_DIRECTORY): os.mkdir(GENERATED_DIRECTORY)
    for bitwidth in BITWIDTHS:
        for mdescr, mkind, extra_args in MKINDS:
            for extra_arg in extra_args:
                for vkind in VKINDS:
                    write_if_different(os.path.join(GENERATED_DIRECTORY, prime_to_fname(prime, bitwidth, mdescr + '_' + vkind + extra_arg, ext='.v')),
                                       r'''Require Import Crypto.Rewriter.PerfTesting.Core.
Global Set Printing Width 1000000.
Goal True.
  %s.perf%s "%s" %s%s.
  exact I.
Defined.
''' % (mkind, vkind, prime, bitwidth, extra_arg))
                write_if_different(os.path.join(GENERATED_DIRECTORY, prime_to_fname(prime, bitwidth, mdescr + '_extracted' + extra_arg, ext='.sh')),
                                   r'''#!/usr/bin/env bash
%s "%s" %s%s
''' % (os.path.join(EXTRACTION_DIRECTORY, 'perf_' + mdescr), prime, bitwidth, extra_arg.replace('%nat', '')))

def write_makefile(primes):
    if not os.path.exists(GENERATED_DIRECTORY): os.mkdir(GENERATED_DIRECTORY)
    vos = ' \\\n\t'.join('src/Rewriter/PerfTesting/Specific/generated/%s' % prime_to_fname(prime, bitwidth, mdescr + '_' + vkind + extra_arg, ext='.vo')
                         for prime in primes
                         for bitwidth in BITWIDTHS
                         for mdescr, mkind, extra_args in MKINDS
                         for extra_arg in extra_args
                         for vkind in VKINDS)
    shs = ' \\\n\t'.join('src/Rewriter/PerfTesting/Specific/generated/%s' % prime_to_fname(prime, bitwidth, mdescr + '_extracted' + extra_arg, ext='.sh')
                         for prime in primes
                         for bitwidth in BITWIDTHS
                         for mdescr, mkind, extra_args in MKINDS
                         for extra_arg in extra_args)

    # We unconditionally write to primes.mk, so that invoking this target actually updates the timestamp
    with open(os.path.join(GENERATED_DIRECTORY, 'primes.mk'), 'w') as f:
        f.write(r'''PERF_PRIME_VOS := %s
PERF_PRIME_SHS := %s
''' % (vos, shs))

if __name__ == '__main__':
    primes = tuple(filter_lines(get_input(sys.argv[1])))
    for prime in primes: write_files(prime)
    write_makefile(primes)




# from http://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-/

__all__ = ["memoize"]

def memoize(f):
    """ Memoization decorator for a function taking one or more arguments. """
    class memodict(dict):
        def __getitem__(self, *key, **kwkey):
            return dict.__getitem__(self, (tuple(key), tuple((k, kwkey[k]) for k in sorted(kwkey.keys()))))

        def __missing__(self, key):
            args, kwargs = key
            self[key] = ret = f(*args, **dict(kwargs))
            return ret

    return memodict().__getitem__



#!/usr/bin/env python
from __future__ import with_statement
from memoize import memoize
import codecs, re, sys
import random

LAMBDA = u'\u03bb'

OP_NAMES = {'*':'MUL', '+':'ADD', '>>':'SHL', '<<':'SHR', '|':'OR', '&':'AND'}

MAX_INSTRUCTION_WINDOW = 1000

INSTRUCTIONS_PER_CYCLE = 4

REGISTERS = tuple(['r%d' % i for i in range(8, 16)]
                  + ['RAX', 'RBX', 'RCX', 'RDX', 'RSI', 'RDI', 'RBP', 'RSP'])

CORE_DATA = tuple(('p%d' % i, 1) for i in range(8))
CORES = tuple(name for name, count in CORE_DATA)
CORE_COUNT = dict(CORE_DATA)

def possible_cores_for_line(line, var_types):
    REGISTER = 'reg'
    MEMORY = 'mem'
    IMMEDIATE = 'imm'
    VECTOR = 'vec'
    INPUT = 'input'
    def arg_type(arg):
        if arg in REGISTERS: return REGISTER
        if arg[:2] == '0x' and arg[2:].isdigit(): return IMMEDIATE
        if arg[:3] in ('mmx', 'xmm', 'ymm'): return VECTOR
        if arg[0] in 'v': return VECTOR
        if arg[0] in 'Rr': return REGISTER
        if arg[0] in 'm': return MEMORY
        if arg[0] in 'x': return INPUT
        assert False
    # from page 233 of http://agner.org/optimize/instruction_tables.pdf
    if line['op'] == '*':
        if line['type'] == 'uint64_t' and '0x13' in line['args']: # * 19 can be either imul r64/r64/i, or two lea; we skip the second case because jgross can't figure out what cost to use for it
            return ({
                'core': ({ 'name' : 'p1' , 'latency' : 1 },),
                'latency' : 3,
                'instruction' : 'IMUL r64,r64,i'
                },)
        elif line['type'] == 'uint128_t' and all(var_types[var] == 'uint64_t' for var in line['args']): # mulx
            return ({
                'core': tuple({ 'name' : core_name , 'latency' : 1 } for core_name in ('p1', 'p5')),
                'latency' : 4,
                'instruction' : 'MULX r64,r64,r64'
                },)
        else:
            assert False
    elif line['op'] == '+':
        if line['type'] == 'uint128_t':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1+1 },),
                'latency' : 1+1,
                'instruction' : 'ADD; ADC(X)'
                } for core_name in ('p0', 'p6'))
        elif line['type'] == 'uint64_t':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1 },),
                'latency' : 1,
                'instruction' : 'ADD'
                } for core_name in ('p0', 'p1', 'p5', 'p6'))
        else:
            assert False
    elif line['op'] in ('>>', '<<'):
        if var_types[line['args'][0]] == 'uint128_t' and line['type'] == 'uint64_t' and line['args'][1][:2] == '0x':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1 },),
                'latency' : 3,
                'instruction' : ('SHLD' if line['op'] == '<<' else 'SHRD') + ' r,r,i'
                } for core_name in ('p1',))
        elif var_types[line['args'][0]] == 'uint64_t' and line['type'] == 'uint64_t' and line['args'][1][:2] == '0x':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1 },),
                'latency' : 1,
                'instruction' : ('SHL' if line['op'] == '<<' else 'SHR') + ' r,i'
                } for core_name in ('p0', 'p6'))
        else:
            assert False
    elif line['op'] in ('&', '|', '^'):
        return tuple({
            'core' : ({ 'name' : core_name , 'latency' : 1 },),
            'latency' : 1,
            'instruction' : {'&':'AND', '|':'OR', '^':'XOR'}[line['op']]
            } for core_name in ('p0', 'p1', 'p5', 'p6'))
    elif line['op'] in ('LOAD','MOVmr') or (line['op'] == '=' and arg_type(line['args'][1]) in (MEMORY,) and arg_type(line['args'][0]) in (REGISTER,)):
        if line['type'] == 'uint128_t': # issue 2 MOV, same port, block on p4
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 2 }, { 'name' : 'p4' , 'latency' : 2 }),
                'latency' : 2,
                'instruction' : 'MOV m,r; MOV m,r'
                } for core_name in ('p2', 'p3', 'p7'))
        elif line['type'] == 'uint64_t':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1 }, { 'name' : 'p4' , 'latency' : 1 }),
                'latency' : 1,
                'instruction' : 'MOV m,r'
                } for core_name in ('p2', 'p3', 'p7'))
        else:
            assert False
    elif line['op'] in ('STORE','MOVrm') or (line['op'] == '=' and arg_type(line['args'][0]) in (MEMORY,) and arg_type(line['args'][1]) in (REGISTER,)):
        if line['type'] == 'uint128_t': # issue 2 MOV, different ports
            return ({
                'core' : tuple({ 'name' : core_name , 'latency' : 1 } for core_name in ('p2', 'p3')),
                'latency' : 1,
                'instruction' : 'MOV r64,m; MOV r64,m'
                },)
        elif line['type'] == 'uint64_t':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1 },),
                'latency' : 1,
                'instruction' : 'MOV r64,m'
                } for core_name in ('p2', 'p3'))
        else:
            assert False
    elif line['op'] in ('MOV','MOVrr') or (line['op'] == '=' and arg_type(line['args'][1]) in (REGISTER,INPUT) and arg_type(line['args'][0]) in (REGISTER,INPUT)):
        if line['type'] == 'uint64_t':
            return tuple({
                'core' : ({ 'name' : core_name , 'latency' : 1 },),
                'latency' : 1,
                'instruction' : 'MOV r64,r64'
                } for core_name in ('p0', 'p1', 'p5', 'p6'))
        else:
            assert False
    else:
        assert False



if len(sys.argv) > 1:
    MAX_INSTRUCTION_WINDOW = int(sys.argv[1])

def get_lines(filename):
    with codecs.open(filename, 'r', encoding='utf8') as f:
        lines = f.read().replace('\r\n', '\n')
    return [line.strip() for line in re.findall("%s '.*?[Rr]eturn [^\r\n]*" % LAMBDA, lines, flags=re.DOTALL)[0].split('\n')]

def strip_casts(text):
    return re.sub(r'\(u?int[0-9]*_t\)\s*\(?([^\)]*)\)?', r'\1', text)

def get_input_var_names(input_data):
    return tuple(i for i in input_data['vars'].replace('%core', '').replace(',', ' ').replace('(', ' ').replace(')', ' ').replace("'", ' ').split(' ')
                 if i != '')

def parse_lines(lines):
    orig_lines = list(lines)
    lines = list(map(strip_casts, orig_lines))
    assert lines[0][:len(LAMBDA + ' ')] == LAMBDA + ' '
    assert lines[0][-1] == ','
    ret = {}
    ret['vars'] = lines[0][len(LAMBDA + ' '):-1]
    assert lines[-1][-1] == ')'
    ret['return'] = lines[-1][:-1].replace('return ', '').replace('Return ', '')
    ret['header'] = orig_lines[0]
    ret['footer'] = orig_lines[-1]
    ret['lines'] = []
    var_types = dict((var, 'uint64_t') for var in get_input_var_names(ret))
    for line, orig_line in zip(lines, orig_lines)[1:-1]:
        binop = re.findall('^(u?int[0-9]*_t) ([^ ]*) = ([^ ]*) ([^ ]*) ([^ ]*);(?: // .*)?$', line)
        unop = re.findall('^([^ ]*) = ([^ ]*);(?: // .*)?$', line)
        if len(binop) > 0:
            datatype, varname, arg1, op, arg2 = binop[0]
            var_types[varname] = datatype
            cur_line = {'type':datatype, 'out':varname, 'op':op, 'args':(arg1, arg2), 'source':orig_line}
        else:
            varname, arg = unop[0]
            var_types[varname] = var_types[arg]
            cur_line = {'type':var_types[varname], 'out':varname, 'op':'=', 'args':(arg,), 'source':orig_line}
        possible_cores = possible_cores_for_line(cur_line, var_types)
        cur_line['cores'] = possible_cores
        ret['lines'].append(cur_line)
    print('Compiling %d lines in groups of %d...' % (len(ret['lines']), min(MAX_INSTRUCTION_WINDOW, len(ret['lines']))))
    ret['lines'] = tuple(ret['lines'])
    split_ret = []
    for start in range(0, len(ret['lines']), MAX_INSTRUCTION_WINDOW):
        cur_ret = dict(ret)
        cur_ret['lines'] = ret['lines'][start:][:MAX_INSTRUCTION_WINDOW]
        split_ret.append(cur_ret)
    return tuple(split_ret)

def get_var_names(input_data):
    return tuple(line['out'] for line in input_data['lines'])

def get_output_var_names(input_data):
    return tuple(i for i in data['return'].replace(',', ' ').replace('(', ' ').replace(')', ' ').split(' ')
                 if i != '')

def create_set(name, items):
    ret = ''
    ret += 'set of int: %s = 1..%d;\n' % (name, len(items))
    for i, item in enumerate(items):
        ret += '%s: %s = %d; ' % (name, item, i+1)
    ret += 'array[%s] of string: %s_NAMES = ["' % (name, name)
    ret += '", "'.join(items) + '"];\n'
    ret += '\n'
    return ret

def schedule(data, basepoint, do_print):
    def make_data_dependencies(input_data):
        input_var_names = get_input_var_names(input_data)
        dependencies = dict((var, tuple()) for var in input_var_names)
        for line in input_data['lines']:
            dependencies[line['out']] = tuple(arg for arg in line['args']
                                              if arg[0] not in '0123456789')
        return dependencies

    def make_reverse_data_dependencies(dependencies):
        reverse_dependencies = {}
        for k, v in dependencies.items():
            for arg in v:
                if arg not in reverse_dependencies.keys(): reverse_dependencies[arg] = []
                reverse_dependencies[arg].append(k)
        return reverse_dependencies

    def get_possible_next_statements(remaining_vars, dependencies):
        return [var for var in remaining_vars
                if all(arg not in remaining_vars for arg in dependencies[var])]

    def make_initial_core_state():
        return {'cores':dict((name, [0] * count) for name, count in CORE_DATA),
                'registers':dict((name, None) for name in REGISTERS)}

#    def freeze_core_state(core_state):
#        return (tuple(tuple(core_state['cores'][name]) for name in CORES),
#                tuple(core_state['registers'][name] for name in REGISTERS))
#    def unfreeze_core_state(core_state):
#        return {'cores':dict((name, list(cycles_until_free)) for name, cycles_until_free in zip(CORES, core_state[0])),
#                'registers':dict((name, var) for name, var in zip(REGISTERS, core_state[1]))}

    def get_initial_indices(input_data):
        #if basepoint == 0:
        #    return tuple(list(get_input_var_names(input_data)) + list(get_var_names(input_data)))
        #else:
        return tuple(get_var_names(input_data))



#    def make_source(input_data, var):
#        input_var_names = get_input_var_names(input_data)
#        if var in input_var_names: return 'LOAD %s' % var

#    def freeze_core_state(vars_ready_at, core_remaining_cycle_count,



    dependencies = make_data_dependencies(data)
    reverse_dependencies = make_reverse_data_dependencies(dependencies)

    def make_initial_core_state():
        vars_remaining_cycles = {}
        core_remaining_cycle_count = dict([(core_name, [0] * core_count) for core_name, core_count in CORE_DATA]
                                          + [(core_name, [0]) for core_name in REGISTERS])
        cur_instructions_in_cycle = 0
        register_vals = dict((var, None) for var in REGISTERS)
        return vars_remaining_cycles, core_remaining_cycle_count, cur_instructions_in_cycle, register_vals

    def freeze_gen(v, rec=(lambda x:x)):
        if isinstance(v, list):
            return ('list', tuple(map(rec, v)))
        if isinstance(v, tuple):
            return ('tuple', tuple(map(rec, v)))
        elif isinstance(v, dict):
            return ('dict', tuple(sorted((k, rec(val)) for k, val in v.items())))
        else:
            return v
    def unfreeze_gen(v, rec=(lambda x:x)):
        if isinstance(v, tuple):
            ty, v = v
            if ty == 'list':
                return list(map(rec, v))
            elif ty == 'tuple':
                return tuple(map(rec, v))
            elif ty == 'dict':
                return dict((k, rec(val)) for k, val in v)
            else:
                print('Freeze error: %s' % repr((ty, v)))
                assert False
        else:
            return v
    def freeze(v):
        return freeze_gen(v, freeze)
    def unfreeze(v):
        return unfreeze_gen(v, unfreeze)


    def is_mul(core, use_imul=False):
        return 'MULX' in core['instruction'] or (use_imul and 'IMUL' in core['instruction'])

    def get_mulx_rdx_args(core, args, use_imul=False):
        if is_mul(core, use_imul=use_imul):
            if args[0][:2] == '0x': return (args[1],)
            return tuple(args[:1])
            return tuple(sorted(args, key=(lambda x: int(x.lstrip('0xrs'))))[:1])
        return tuple()

    def update_register_vals_with_core_args(var, core, args, register_vals):
        changed = False
        rdx_is_last_imul_output = False
        if 'last_imul_output' not in register_vals: register_vals['last_imul_output'] = None
        for new_rdx in get_mulx_rdx_args(core, args, use_imul=True):
            rdx_is_last_imul_output = (register_vals['last_imul_output'] == new_rdx)
            changed = (register_vals['RDX'] != new_rdx)
            register_vals['RDX'] = new_rdx
            if 'IMUL' in core['instruction']: register_vals['last_imul_output'] = var
        return changed, rdx_is_last_imul_output, register_vals

    @memoize
    def update_core_state(var, core, args, core_state):
        core = unfreeze(core)
        (vars_remaining_cycles, core_remaining_cycle_count, cur_instructions_in_cycle, register_vals) = unfreeze(core_state)
        changed, rdx_is_last_imul_output, register_vals = update_register_vals_with_core_args(var, core, args, register_vals)
        cost = 0
        if cur_instructions_in_cycle >= INSTRUCTIONS_PER_CYCLE:
            cost += 1
            cur_instructions_in_cycle = 0
            for c in core_remaining_cycle_count.keys():
                for i in range(len(core_remaining_cycle_count[c])):
                    core_remaining_cycle_count[c][i] = max(0, core_remaining_cycle_count[c][i] - 1)
            vars_remaining_cycles = dict((var, c - 1) for var, c in vars_remaining_cycles.items()
                                         if c > 1)
        cycles_passed = max([min(core_remaining_cycle_count[port['name']]) for port in core['core']] +
                            [vars_remaining_cycles[v] for v in dependencies[var] if v in vars_remaining_cycles.keys()])
        if cycles_passed != 0:
            cost += cycles_passed
            cur_instructions_in_cycle = 1
            for c in core_remaining_cycle_count.keys():
                for i in range(len(core_remaining_cycle_count[c])):
                    core_remaining_cycle_count[c][i] = max(0, core_remaining_cycle_count[c][i] - cycles_passed)
            vars_remaining_cycles = dict((var, c - cycles_passed) for var, c in vars_remaining_cycles.items()
                                         if c > cycles_passed)
        else:
            cur_instructions_in_cycle += 1
        vars_remaining_cycles[var] = core['latency']
        for port in core['core']:
            assert core_remaining_cycle_count[port['name']][0] == 0
            core_remaining_cycle_count[port['name']][0] = port['latency']
            core_remaining_cycle_count[port['name']] = sorted(core_remaining_cycle_count[port['name']])
        return (cost, freeze((vars_remaining_cycles, core_remaining_cycle_count, cur_instructions_in_cycle, register_vals)))

    @memoize
    def evaluate_cost_memo(arg):
        schedule, core_state = unfreeze_gen(arg)
        schedule = unfreeze(schedule)
        (vars_remaining_cycles, core_remaining_cycle_count, cur_instructions_in_cycle, register_vals) = unfreeze(core_state)
        if len(schedule) == 0: return max([0] + list(vars_remaining_cycles.values()))
        (var, core, args), schedule = schedule[0], schedule[1:]
        cost, core_state = update_core_state(var, freeze(core), args, core_state)
        return cost + evaluate_cost_memo(freeze_gen((freeze(schedule), core_state)))


    def get_wait_times(var_cores, core_state):
        for var, core, args in var_cores:
            (vars_remaining_cycles, core_remaining_cycle_count, cur_instructions_in_cycle, register_vals) = unfreeze(core_state)
            changed, rdx_is_last_imul_output, register_vals = update_register_vals_with_core_args(var, core, args, register_vals)
            cost, new_core_state = update_core_state(var, freeze(core), args, core_state)
            yield (cost, -len(reverse_dependencies.get(var, [])), changed, -core['latency'], not rdx_is_last_imul_output, var, get_mulx_rdx_args(core, args, use_imul=True), core, args, new_core_state)

    def cmp_wait_time(v1, v2):
        (cost1, neg_len_deps1, changed1, neg_latency1, not_rdx_is_last_imul_output1, var1, mulx_rdx_args, core1, args1, new_core_state1) = v1
        (cost2, neg_len_deps2, changed2, neg_latency2, not_rdx_is_last_imul_output2, var2, mulx_rdx_args, core2, args2, new_core_state2) = v2
        if cost1 != cost2: return cmp(cost1, cost2)
        if core1['instruction'] == core2['instruction'] or (is_mul(core1, use_imul=True) and is_mul(core2, use_imul=True)):
            if changed1 != changed2: return cmp(changed1, changed2)
            if not_rdx_is_last_imul_output1 != not_rdx_is_last_imul_output2: return cmp(not_rdx_is_last_imul_output1, not_rdx_is_last_imul_output2)
            if core1['instruction'] != core2['instruction'] and \
               (True or get_mulx_rdx_args(core1, args1, use_imul=True) == get_mulx_rdx_args(core2, args2, use_imul=True)):
                return cmp(var1, var2)
            if neg_len_deps1 != neg_len_deps2: return cmp(neg_len_deps1, neg_len_deps2)
            if neg_latency1 != neg_latency2: return cmp(neg_latency1, neg_latency2)
        if var1 != var2: return cmp(var1, var2)
        return cmp(v1, v2)

    def get_sorted_next_statements(next_var_cores, core_state):
        return sorted(get_wait_times(next_var_cores, core_state), cmp=cmp_wait_time)

    def add_cycle_info(schedule):
        core_state = freeze(make_initial_core_state())
        schedule_with_cycle_info = []
        cur_cycle = 0
        for var, core, args in schedule:
            cost, core_state = update_core_state(var, freeze(core), args, core_state)
            cur_cycle += cost
            schedule_with_cycle_info.append((var,
                                             {'start':cur_cycle, 'finish':cur_cycle + core['latency']},
                                             core,
                                             args))
        return schedule_with_cycle_info

    def evaluate_cost(schedule_with_cycle_info):
        return max(cycles['finish'] for var, cycles, core in reversed(schedule_with_cycle_info))


    var_names = get_var_names(data)
    lines = dict((line['out'], line) for line in data['lines'])

    @memoize
    def schedule_remaining(remaining_indices, core_state):
        def make_schedule(var, core, args):
            cost, new_core_state = update_core_state(var, freeze(core), args, core_state)
            extra_cost, schedule = schedule_remaining(tuple(i for i in remaining_indices if i != var), new_core_state)
            return cost + extra_cost, ([(var, core, args)] + schedule)
        next_statements = get_possible_next_statements(remaining_indices, dependencies)
        min_cost, min_schedule = None, None
        var_cores = [(var, core, (lines[var]['args'] if var in lines.keys() else (var,)))
                     for var in next_statements
                     for core in (lines[var]['cores'] if var in lines.keys() else possible_cores_for_line({'op':'LOAD', 'type':'uint64_t'}))]

        sorted_subset_next_statements = sorted_next_statements = get_sorted_next_statements(var_cores, core_state)
        if len(sorted_next_statements) > 0:
            pre_min_cost = sorted_next_statements[0][0]
#            print((pre_min_cost, tuple(var for cost2, var, core, new_core_state in sorted_next_statements if pre_min_cost == cost2)))
            rdx_with_imul = set(arg
                                for cost, reverse_dep_count, changed, neg_latency, not_rdx_is_last_imul_output, var, mulx_rdx_args, core, args, new_core_state in sorted_next_statements
                                for arg in mulx_rdx_args
                                if pre_min_cost == cost and 'IMUL' in core['instruction'])
            sorted_subset_next_statements \
                = tuple((cost, var, core, args, new_core_state, mulx_rdx_args) for cost, reverse_dep_count, changed, neg_latency, not_rdx_is_last_imul_output, var, mulx_rdx_args, core, args, new_core_state in sorted_next_statements
                        if pre_min_cost == cost)
            if False and any(all(arg not in rdx_with_imul for arg in mulx_rdx_args)
                   for cost, var, core, args, new_core_state, mulx_rdx_args in sorted_subset_next_statements
                   if len(mulx_rdx_args) > 0):
                sorted_subset_next_statements = tuple([(cost, var, core, args, new_core_state, mulx_rdx_args)
                                                       for cost, var, core, args, new_core_state, mulx_rdx_args in sorted_subset_next_statements
                                                       if all(arg not in rdx_with_imul for arg in mulx_rdx_args)]
                                                      + [(cost, var, core, args, new_core_state, mulx_rdx_args)
                                                         for cost, var, core, args, new_core_state, mulx_rdx_args in sorted_subset_next_statements
                                                         if not all(arg not in rdx_with_imul for arg in mulx_rdx_args)])
            sorted_subset_next_statements = sorted_subset_next_statements[:1]
            if pre_min_cost == 0: sorted_subset_next_statements = sorted_subset_next_statements[:1]
        for cost, var, core, args, new_core_state, mulx_rdx_args in sorted_subset_next_statements:
            cost, schedule = make_schedule(var, core, args)
            if min_cost is None or cost < min_cost:
                min_cost, min_schedule = cost, schedule
#            return min_cost, min_schedule
        if min_cost is None:
            min_cost, min_schedule = evaluate_cost_memo(freeze_gen((freeze([]), core_state))), []
        return min_cost, min_schedule

    def get_live_ranges(var_to_line, schedule_with_cycles, RET_loc):
        var_names = get_var_names(data)
        input_var_names = get_input_var_names(data)
        output_var_names = get_output_var_names(data)
        ret = dict((var, {'start':0, 'accessed':[], 'mul_accessed':[]}) for var in input_var_names)
        for (var, locs, core, args) in schedule_with_cycles:
            assert var not in ret.keys()
            line = var_to_line[var]
            ret[var] = {'start':locs['start'], 'accessed':[], 'mul_accessed':[]}
            for arg in line['args']:
                if arg in var_names + input_var_names:
                    for latency in range(max(c['latency'] for c in core['core'])): # handle instructions that need data for multiple cycles, like add;adcx
                        ret[arg]['accessed'].append(locs['start'] + latency)
                elif arg[:2] == '0x':
                    pass
                else:
                    print(arg)
            for arg in get_mulx_rdx_args(core, args, use_imul=True):
                ret[arg]['mul_accessed'].append(locs['start'])
        for var in output_var_names:
            ret[var]['accessed'].append(RET_loc)
        for var in ret.keys():
            ret[var]['end'] = max(ret[var]['accessed'])
        for var in ret.keys():
            if not (len(ret[var]['mul_accessed']) == 0 or tuple(ret[var]['mul_accessed']) == tuple(ret[var]['accessed'])):
                print((var, ret[var]['accessed'], ret[var]['mul_accessed']))
                assert False
        return ret

    def remake_overlaps(live_ranges):
        live_ranges = dict(live_ranges)
        for var in live_ranges.keys():
            live_ranges[var] = dict(live_ranges[var])
            live_ranges[var]['end'] = max(live_ranges[var]['accessed'])
            live_ranges[var]['overlaps'] = tuple(sorted(
                other_var
                for other_var in live_ranges.keys()
                if other_var != var and
                (live_ranges[other_var]['start'] <= live_ranges[var]['start'] <= live_ranges[other_var]['end']
                 or live_ranges[var]['start'] <= live_ranges[other_var]['start'] <= live_ranges[var]['end'])
                ))
        return live_ranges

    def make_initial_register_ranges(RET_loc):
        return dict((reg, [None] * (RET_loc + 2)) for reg in REGISTERS)

    def force_mulx_args(live_ranges, register_ranges):
        for var in live_ranges.keys():
            for loc in live_ranges[var]['mul_accessed']:
                assert register_ranges['RDX'][loc] is None
                register_ranges['RDX'][loc] = var
        return register_ranges

    def lookup_var_in_reg(register_ranges, loc, var):
        var_dict = dict((register_ranges[reg][loc], reg) for reg in register_ranges.keys())
        next_var_dict = dict((register_ranges[reg][loc+1], reg) for reg in register_ranges.keys())
        next_next_var_dict = dict((register_ranges[reg][loc+2], reg) for reg in register_ranges.keys())
        if var in var_dict.keys():
            return [var_dict[var]]
        elif var + '_low' in var_dict.keys() and var + '_high' in var_dict.keys():
            return ['%s:%s' % (var_dict[var + '_high'], var_dict[var + '_low'])]
        elif var + '_low' in var_dict.keys() and var + '_high' in next_var_dict.keys():
            return ['%s:%s' % (next_var_dict[var + '_high'], var_dict[var + '_low'])]
        elif var + '_low' in next_var_dict.keys() and var + '_high' in next_next_var_dict.keys():
            return ['%s:%s' % (next_next_var_dict[var + '_high'], next_var_dict[var + '_low'])]
        else:
            return []

    def update_source(line, loc, register_ranges):
        source = line['source']
        if line['op'] in ('&', '|', '^') and line['type'] == 'uint64_t' and line['args'][1][:2] == '0x':
            for reg in lookup_var_in_reg(register_ranges, loc, line['out']):
                source = source.replace(line['out'], reg)
            for reg in lookup_var_in_reg(register_ranges, loc, line['args'][0] + '_low'):
                source = source.replace(line['args'][0], reg)
            for reg in lookup_var_in_reg(register_ranges, loc, line['args'][0]):
                source = source.replace(line['args'][0], reg)
        else:
            for var in sorted([line['out']] + list(line['args']), key=len):
                for reg in lookup_var_in_reg(register_ranges, loc, var):
                    source = source.replace(var, reg)
        return source

    def get_next_registers_use(loc, register_ranges, live_ranges):
        def gen():
            for reg in REGISTERS:
                last_vals = [val for val in register_ranges[reg][:loc] if val is not None]
                if len(last_vals) == 0:
                    yield (reg, None)
                else:
                    val = last_vals[-1].replace('_low', '').replace('_high', '')
                    accessed = [aloc for aloc in live_ranges[val]['accessed'] if aloc > loc]
                    if len(accessed) == 0:
                        yield (reg, None)
                    else:
                        yield (reg, accessed[0])
        return dict(gen())

    def free_registers(loc, register_ranges, live_ranges):
        all_free_registers = [reg for reg in REGISTERS if register_ranges[reg][loc] is None]
        if loc == 0: return tuple(all_free_registers)
        if loc == 1:
            return tuple([reg for reg in all_free_registers if register_ranges[reg][loc-1] is None]
                         + [reg for reg in all_free_registers if register_ranges[reg][loc-1] is not None])
        next_uses = get_next_registers_use(loc, register_ranges, live_ranges)
        def get_key(reg):
            if register_ranges[reg][loc-1] is None and register_ranges[reg][loc-2] is None:
                return (0, next_uses[reg])
            if register_ranges[reg][loc-1] is None and register_ranges[reg][loc-2] is not None:
                return (1, next_uses[reg])
            return (2, next_uses[reg])
        return tuple(sorted(all_free_registers, key=get_key))

    def spill_register(loc, register_ranges, exclude_vals):
        # kick out the value that will disappear soonest
        empties = dict((reg, [new_loc for new_loc, val in enumerate(register_ranges[reg][loc:]) if val is None])
                       for reg in register_ranges.keys()
                       if reg != 'RDX' and register_ranges[reg][loc].replace('_low', '').replace('_high', '') not in exclude_vals)
        sorted_empties = sorted((new_locs[0], reg) for reg, new_locs in empties.items() if len(new_locs) > 0)
        cycles, reg = sorted_empties[0]
        print('Spilling %s from %s!' % (register_ranges[reg][loc], reg))
        for new_loc in range(loc, loc + cycles):
            register_ranges[reg][new_loc] = None
        return reg, register_ranges

    def prune_stores(stores_available, movs_needed):
        movs_used = set(arg for mov_type, reg, arg, loc in movs_needed)
        for mov_type, arg, reg, loc in stores_available:
            if arg in movs_used:
                yield (mov_type, arg, reg, loc)

    def insert_movs(schedule_with_cycles, register_ranges, live_registers):
        pass

    def linear_allocate(var_to_line, schedule_with_cycles, live_ranges, register_ranges):
        movs_needed = []
        stores_available = []
        for (var, locs, core, args) in schedule_with_cycles:
            registers = free_registers(locs['start'], register_ranges, live_ranges)
            line = var_to_line[var]
#            print((len(registers), line['source'], tuple((reg, register_ranges[reg][locs['start']]) for reg in REGISTERS)))
            if line['type'] == 'uint128_t' and line['op'] == '+':
                for latency, bits in ((0, '_low'), (1, '_high')):
                    for argi, arg in enumerate(line['args']):
                        found = list(lookup_var_in_reg(register_ranges, locs['start'] + latency, arg + bits))
                        if len(found) == 0:
                            if len(registers) == 0:
                                reg, register_ranges = spill_register(locs['start'] + latency, register_ranges, line['args'])
                            else:
                                reg, registers = registers[0], registers[1:]
                            assert register_ranges[reg][locs['start'] + latency] is None
                            register_ranges[reg][locs['start'] + latency] = arg + bits
                            movs_needed.append(('MOVmr', arg + bits, reg, locs['start'] + latency))
                        else:
                            reg = found[0]
                            if reg == 'RDX':
                                if locs['start'] + latency == 0:
                                    movs_needed.append(('MOVmr', arg + bits, reg, locs['start'] + latency))
                                elif register_ranges['RDX'][locs['start'] + latency - 1] != arg + bits:
                                    # MOVrr if the value is in some register, otherwise (if the filtered list is empty), MOVmr
                                    mov_type, from_val = ([('MOVrr', reg_from) for reg_from in register_ranges.keys()
                                                           if register_ranges['RDX'][locs['start'] + latency - 1] == arg + bits]
                                                          + [('MOVmr', arg + bits)])[0]
                                    movs_needed.append((mov_type, from_val, reg, locs['start'] + latency))
                        if argi == 0:
                            register_ranges[reg][locs['start'] + latency + 1] = line['out'] + bits
                            stores_available.append(('MOVrm', reg, line['out'] + bits, locs['start'] + latency + 1))
            elif line['type'] == 'uint64_t' and line['op'] == '>>' and line['args'][0] in var_to_line.keys() and var_to_line[line['args'][0]]['type'] == 'uint128_t' and line['args'][1][:2] == '0x':
                for arg, latency in ((line['out'], core['latency']), (line['args'][0] + '_low', 0), (line['args'][0] + '_high', 0)):
                    found = list(lookup_var_in_reg(register_ranges, locs['start'], arg))
                    if len(found) == 0:
                        if len(registers) == 0:
                            reg, register_ranges = spill_register(locs['start'] + latency, register_ranges, line['args'])
                        else:
                            reg, registers = registers[0], registers[1:]
                        assert register_ranges[reg][locs['start'] + latency] is None
                        register_ranges[reg][locs['start'] + latency] = arg
                        if arg != line['out']:
                            for c in range(latency):
                                movs_needed.append(('MOVmr', arg, reg, locs['start'] + c))
                        else:
                            stores_available.append(('MOVrm', reg, line['out'] + bits, locs['start'] + latency))
            else:
                if line['type'] == 'uint128_t':
                    out_args = [line['out'] + '_high', line['out'] + '_low']
                else:
                    out_args = [line['out']]
                for arg in sorted(out_args + list(line['args']), key=len):
                    if arg[:2] == '0x': continue
                    if line['type'] == 'uint64_t' and arg in var_to_line.keys() and var_to_line[arg]['type'] == 'uint128_t':
                        if line['op'] in ('&', '|', '^'): arg = arg + '_low'
                    found = list(lookup_var_in_reg(register_ranges, locs['start'], arg))
                    if len(found) == 0:
                        if len(registers) == 0:
                            reg, register_ranges = spill_register(locs['start'], register_ranges, out_args + list(line['args']))
                        else:
                            reg, registers = registers[0], registers[1:]
                        assert register_ranges[reg][locs['start']] is None
                        if arg in out_args:
                            for latency in range(core['latency']+1):
                                register_ranges[reg][locs['start'] + latency] = arg
                            stores_available.append(('MOVrm', reg, arg, locs['start'] + core['latency']))
                        else:
                            for latency in range(max(c['latency'] for c in core['core'])): # handle instructions that need data for multiple cycles, like add;adcx
                                register_ranges[reg][locs['start'] + latency] = arg
                            movs_needed.append(('MOVmr', arg, reg, locs['start']))
                    elif arg in out_args:
                        assert False
                    else:
                        reg = found[0]
                        if reg == 'RDX':
                            if locs['start'] == 0:
                                movs_needed.append(('MOVmr', arg, reg, locs['start']))
                            elif register_ranges['RDX'][locs['start'] - 1] != arg:
                                # MOVrr if the value is in some register, otherwise (if the filtered list is empty), MOVmr
                                mov_type, from_val = ([('MOVrr', reg_from) for reg_from in register_ranges.keys()
                                                       if register_ranges['RDX'][locs['start'] - 1] == arg]
                                                      + [('MOVmr', arg)])[0]
                                movs_needed.append((mov_type, from_val, reg, locs['start']))
                        
            print(var_to_line[var]['source'] + ' // ' + update_source(var_to_line[var], locs['start'], register_ranges))
        movs_needed = sorted(movs_needed, key=(lambda v: v[3]))
        stores_needed = sorted(prune_stores(stores_available, movs_needed), key=(lambda v: v[3]))
        print(len(movs_needed))
        print(len(stores_needed))
        for i in movs_needed: print(i)
        for i in stores_needed: print(i)
#        sys.exit(0)

#    def insert_possible_registers(live_ranges):
#        live_ranges = dict(live_ranges)
#        for var in live_ranges.keys():
#            live_ranges[var] = dict(live_ranges[var])
#            live_ranges[var]['possible registers'] = tuple(sorted(
#                other_var
#                for other_var in live_ranges.keys()
#                if other_var != var and
#                (live_ranges[other_var]['start'] <= live_ranges[var]['start'] <= live_ranges[other_var]['end']
##                 or live_ranges[var]['start'] <= live_ranges[other_var]['start'] <= live_ranges[var]['end'])
##                ))
##        return live_ranges

#    def register_allocate(live_ranges):
#        allocated = {}
#        remaining_registers = list(REGISTERS)


    core_state = freeze(make_initial_core_state())
    cost, schedule = schedule_remaining(get_initial_indices(data), core_state) #, freeze_core_state(make_initial_core_state()))
    schedule_with_cycle_info = add_cycle_info(schedule)
    live_ranges = remake_overlaps(get_live_ranges(lines, schedule_with_cycle_info, cost))
    register_ranges = make_initial_register_ranges(cost)
    register_ranges = force_mulx_args(live_ranges, register_ranges)
    print(linear_allocate(lines, schedule_with_cycle_info, live_ranges, register_ranges))
#    print(live_ranges)
#    print(register_ranges)
#    sys.exit(0)
    for var, cycles, core, args in schedule_with_cycle_info:
        if var in lines.keys():
            do_print(lines[var]['source'], ' // %s, start: %s, end: %s' % (core['instruction'], basepoint + cycles['start'], basepoint + cycles['finish']))
        else:
            do_print('%s = %s;' % (var, var), ' // %s, start: %s, end: %s' % (core['instruction'], basepoint + cycles['start'], basepoint + cycles['finish']))
    return basepoint + cost

data_list = parse_lines(get_lines('femulDisplay.log'))
basepoint = 0
for i, data in enumerate(data_list):
    with codecs.open('femulScheduled.log', 'w', encoding='utf8') as f:
        def do_print(v, comment):
            print(v + comment)
            f.write(v + comment + '\n')
        f.write(data['header'] + '\n')
        basepoint = schedule(data, basepoint, do_print)
        f.write(data['footer'] + '\n')
    print(basepoint)
    break



#!/usr/bin/env python3
# USAGE: ./fiat-amd64/gentest.py fiat-amd64/*.asm | while IFS= read -r line; do eval $line 2>/dev/null > /dev/null && echo "$line" || echo "# $line" ; done
import shlex
import sys
import math
import re
import os
from collections import OrderedDict


# in Python 3.9 and newer, this is primitive as str.removesuffix
def removesuffix(s, suffix):
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


def removeprefix(s, prefix):
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


# grep -oP "src/.*square" fiat-c/src/*64*.c

# unsaturated_solinas
solinasprimes = dict(  # num limbs, prime
    curve25519=("5", "2^255 - 19"),
    p448_solinas=("8", "2^448 - 2^224 - 1"),
    p521=("9", "2^521 - 1"),
    poly1305=("3", "2^130 - 5"),
)

montgomeryprimes = dict(
    curve25519_scalar="2^252 + 27742317777372353535851937790883648493",
    p224="2^224 - 2^96 + 1",
    p256="2^256 - 2^224 + 2^192 + 2^96 - 1",
    p256_scalar="2^256 - 2^224 + 2^192 - 89188191075325690597107910205041859247",
    p384="2^384 - 2^128 - 2^96 + 2^32 - 1",
    p384_scalar="2^384 - 1388124618062372383947042015309946732620727252194336364173",
    p434="2^216 * 3^137 - 1",
    secp256k1_montgomery="2^256 - 2^32 - 977",
    secp256k1_scalar="2^256 - 432420386565659656852420866394968145599",
)

saturatedsolinasprimes = dict(curve25519_solinas=("2^255 - 19"))

dettmanprimes = dict(  # last limb width, limbs, last reduction, prime
    secp256k1_dettman=("48", "5", "2", "2^256 - 4294968273")
)

output_makefile = "--makefile" in sys.argv[1:]
directories = tuple(i for i in sys.argv[1:] if i not in ("--makefile",))

asm_op_names = OrderedDict()

regex = re.compile(
    r"fiat_(?P<name>[^_]+(_(solinas|montgomery|dettman))?)_(?P<op>(carry_)?(square|mul|from_bytes|to_bytes|add|sub|opp))"
)
for dirname in directories:
    m = regex.match(os.path.basename(dirname))
    if m:
        for fname in os.listdir(dirname):
            groups = m.groupdict()
            asm_op_names.setdefault((groups["name"], groups["op"]), []).append(
                os.path.join(dirname, fname)
            )


def asm_op_names_key(val):
    (name, op), fnames = val
    n = 0

    if name in dettmanprimes.keys():
        kind = 0
        limbwidth, n, last_reduction, prime = dettmanprimes[name]

    elif name in saturatedsolinasprimes.keys():
        kind = 1
        prime = saturatedsolinasprimes[name]

    elif name in montgomeryprimes.keys():
        kind = 2
        prime = montgomeryprimes[name]

    elif name in solinasprimes.keys():
        kind = 3
        n, prime = solinasprimes[name]

    else:
        assert False, name

    return (kind, n, prime, op, name, fnames)


def is_small(val):
    (kind, n, prime, op, name, fnames) = asm_op_names_key(val)
    prime = eval(prime.replace("^", "**"))
    return math.log2(prime) / 64 <= 4


asm_op_names_items = tuple(sorted(asm_op_names.items(), key=asm_op_names_key))
small_asm_op_names_items = tuple(val for val in asm_op_names_items if is_small(val))

status_file_stems = [
    f"fiat-amd64/{name}-{op}" for (name, op), _fnames in asm_op_names_items
]
small_status_file_stems = [
    f"fiat-amd64/{name}-{op}" for (name, op), _fnames in small_asm_op_names_items
]

status_files = [stem + ".status" for stem in status_file_stems]
only_status_files = [stem + ".only-status" for stem in status_file_stems]
small_status_files = [stem + ".status" for stem in small_status_file_stems]
small_only_status_files = [stem + ".only-status" for stem in small_status_file_stems]

if output_makefile:
    print(
        f"""

# Allow SLOWEST_FIRST=1 to be passed to test files in reverse order.
# When testing interactively, we probably want to test quicker files
# first, to be more snappy.  But on CI, testing the slowest files
# first probably allows for slightly better scheduling when there's
# more parallelism.

AMD64_ASM_STATUS_FILES := $(if $(SLOWEST_FIRST),{' '.join(reversed(status_files))},{' '.join(status_files)})
AMD64_ASM_ONLY_STATUS_FILES := $(if $(SLOWEST_FIRST),{' '.join(reversed(only_status_files))},{' '.join(only_status_files)})
AMD64_ASM_SMALL_STATUS_FILES := $(if $(SLOWEST_FIRST),{' '.join(reversed(small_status_files))},{' '.join(small_status_files)})
AMD64_ASM_SMALL_ONLY_STATUS_FILES := $(if $(SLOWEST_FIRST),{' '.join(reversed(small_only_status_files))},{' '.join(small_only_status_files)})

"""
    )

for item in asm_op_names_items:
    (kind, n, prime, op, name, fnames) = asm_op_names_key(item)
    if kind == 0:
        binary = "src/ExtractionOCaml/dettman_multiplication"
        binary_descr = "Dettman Multiplication"
        limbwidth, _n, last_reduction, _prime = dettmanprimes[name]
        invocation = " ".join(
            [
                binary,
                name,
                "64",
                n,
                limbwidth,
                last_reduction,
                shlex.quote(prime),
                op,
                "--no-wide-int",
                "--shiftr-avoid-uint1",
            ]
            + [
                item
                for fname in fnames
                for item in ("--hints-file", shlex.quote(fname))
            ]
        )
    elif kind == 1:
        binary = "src/ExtractionOCaml/solinas_reduction"
        binary_descr = "Saturated Solinas"
        invocation = " ".join(
            [
                binary,
                name,
                "64",
                shlex.quote(prime),
                op,
                "--no-wide-int",
                "--shiftr-avoid-uint1",
            ]
            + [
                item
                for fname in fnames
                for item in ("--hints-file", shlex.quote(fname))
            ]
        )
    elif kind == 2:
        binary = "src/ExtractionOCaml/word_by_word_montgomery"
        binary_descr = "Word-by-Word Montgomery"
        invocation = " ".join(
            [
                binary,
                name,
                "64",
                shlex.quote(prime),
                op,
                "--no-wide-int",
                "--shiftr-avoid-uint1",
            ]
            + [
                item
                for fname in fnames
                for item in ("--hints-file", shlex.quote(fname))
            ]
        )
    elif kind == 3:
        binary = "src/ExtractionOCaml/unsaturated_solinas"
        binary_descr = "Unsaturated Solinas"
        invocation = " ".join(
            [
                binary,
                name,
                "64",
                n,
                shlex.quote(prime),
                op,
                "--no-wide-int",
                "--shiftr-avoid-uint1",
                "--tight-bounds-mul-by",
                "1.000001",
            ]
            + [
                item
                for fname in fnames
                for item in ("--hints-file", shlex.quote(fname))
            ]
        )
    else:
        assert False, name
    if output_makefile:
        short_fnames = [
            removesuffix(os.path.basename(fname), ".asm") for fname in fnames
        ]
        description = f'{name} {prime.replace(" ", "")} ({op}) ({binary_descr}) ({" ".join(short_fnames)})'
        output_name = f"fiat-amd64/{name}-{op}"
        print(
            f"""
only-test-amd64-files-print-report:: {output_name}.only-status
\t@ test $$(cat $<) -eq 0 || echo 'TEST AMD64 {description} ... \t$(RED)$(BOLD)FAILED$(NORMAL)$(NC)'

test-amd64-files-print-report:: {output_name}.status
\t@ test $$(cat $<) -eq 0 || echo 'TEST AMD64 {description} ... \t$(RED)$(BOLD)FAILED$(NORMAL)$(NC)'

clean::
\t$(HIDE)rm -f {output_name}.status {output_name}.only-status {output_name}.invocation {output_name}.description {output_name}.out {output_name}.out-asm {output_name}.stdout

.PHONY: {output_name}.only-status

{output_name}.status: {binary}

{output_name}.status {output_name}.only-status: {' '.join(fnames)}
\t$(SHOW)'TEST AMD64 {description} ...'
\t$(HIDE)rm -f $@
\t$(HIDE)echo {shlex.quote(invocation + ' -o /dev/null --output-asm /dev/null')} > {output_name}.invocation
\t$(HIDE)echo '{description}' > {output_name}.description
\t$(HIDE)$(TIMER) $(PERF_RECORDER) {invocation} -o {output_name}.out --output-asm {output_name}.out-asm >{output_name}.stdout && \\
\t  {{ echo $$? > $@; echo 'TEST AMD64 {description} ... \t$(GREEN)$(BOLD)PASSED$(NORMAL)$(NC)'; }} || \\
\t  {{ echo $$? > $@; echo 'TEST AMD64 {description} ... \t$(RED)$(BOLD)FAILED$(NORMAL)$(NC)'; \\
\t       echo '================== stdout =================='; \\
\t       cat {output_name}.stdout; \\
\t       echo '============================================'; \\
\t       exit 1; }}
"""
        )
    else:
        print(invocation, "-o", "/dev/null", "--output-asm", "/dev/null")



#!/usr/bin/env python
from __future__ import with_statement
import codecs, re

LAMBDA = u'\u03bb'

MAX_INSTRUCTION_WINDOW = 30

INSTRUCTIONS_PER_CYCLE = 4

REGISTERS = tuple(['RAX', 'RBX', 'RCX', 'RDX', 'RSI', 'RDI', 'RBP', 'RSP']
                  + ['r%d' % i for i in range(8, 16)])

def get_lines(filename):
    with codecs.open(filename, 'r', encoding='utf8') as f:
        lines = f.read().replace('\r\n', '\n')
    return [line.strip() for line in re.findall("%s '.*?[Rr]eturn [^\r\n]*" % LAMBDA, lines, flags=re.DOTALL)[0].split('\n')]

def strip_casts(text):
    return re.sub(r'\(u?int[0-9]*_t\)\s*\(?([^\)]*)\)?', r'\1', text)

def parse_lines(lines):
    lines = list(map(strip_casts, lines))
    assert lines[0][:len(LAMBDA + ' ')] == LAMBDA + ' '
    assert lines[0][-1] == ','
    ret = {}
    ret['vars'] = lines[0][len(LAMBDA + ' '):-1]
    assert lines[-1][-1] == ')'
    ret['return'] = lines[-1][:-1].replace('return ', '').replace('Return ', '')
    ret['lines'] = []
    for line in lines[1:-1]:
        datatype, varname, arg1, op, arg2 = re.findall('^(u?int[0-9]*_t) ([^ ]*) = ([^ ]*) ([^ ]*) ([^ ]*);$', line)[0]
        ret['lines'].append({'type':datatype, 'out':varname, 'op':op, 'args':(arg1, arg2), 'source':line})
    print('Compiling %d lines in groups of %d...' % (len(ret['lines']), min(MAX_INSTRUCTION_WINDOW, len(ret['lines']))))
    ret['lines'] = tuple(ret['lines'])
    split_ret = []
    for start in range(0, len(ret['lines']), MAX_INSTRUCTION_WINDOW):
        cur_ret = dict(ret)
        cur_ret['lines'] = ret['lines'][start:][:MAX_INSTRUCTION_WINDOW]
        split_ret.append(cur_ret)
    return tuple(split_ret)

def get_var_names(input_data):
    return tuple(line['out'] for line in input_data['lines'])

def get_input_var_names(input_data):
    return tuple(i for i in data['vars'].replace('%core', '').replace(',', ' ').replace('(', ' ').replace(')', ' ').replace("'", ' ').split(' ')
                 if i != '')

def get_all_var_names(input_data):
    return tuple(list(get_input_var_names(input_data)) + list(get_var_names(input_data)))

def get_output_var_names(input_data):
    return tuple(i for i in data['return'].replace(',', ' ').replace('(', ' ').replace(')', ' ').split(' ')
                 if i != '')


def line_of_var(input_data, var):
    retv = [line for line in input_data['lines'] if line['out'] == var]
    if len(retv) > 0: return retv[0]
    return {'out': var, 'args':tuple(), 'op': 'INPUT', 'type':'uint64_t'}

def reg_count(size):
    return {'uint64_t':1, 'uint128_t': 2}[size]

def create_set(name, items):
    ret = ''
    ret += 'set of int: %s = 1..%d;\n' % (name, len(items))
    for i, item in enumerate(items):
        ret += '%s: %s = %d; ' % (name, item, i+1)
    ret += 'array[%s] of string: %s_NAMES = ["' % (name, name)
    ret += '", "'.join(items) + '"];\n'
    ret += '\n'
    return ret

######################################################################
######################################################################
##                Assign locations to instructions cumulatively     ##
######################################################################
######################################################################


def make_assign_locations_to_instructions_cumulatively(data):
    def make_decls(input_data):
        var_names = get_all_var_names(input_data)
        ret = ''
        ret += 'include "alldifferent.mzn";\n'
        ret += 'include "cumulative.mzn";\n'
        for line in input_data['lines']:
            ret += '%%%s\n' % line['source']
        ret += create_set('INSTRUCTIONS', list(var_names))
        MAX_NUMBER_OF_NOOPS_PER_INSTRUCTION = 3
        APPROXIMATE_MAX_LATENCY = 6 * INSTRUCTIONS_PER_CYCLE
        max_loc = len(var_names) * MAX_NUMBER_OF_NOOPS_PER_INSTRUCTION + APPROXIMATE_MAX_LATENCY
        ret += 'int: MAX_LOC = %d;\n\n' % max_loc
        ret += 'set of int: LOCATIONS = 1..MAX_LOC;\n'
        ret += 'array[INSTRUCTIONS] of var LOCATIONS: output_locations;\n'
        ret += 'array[INSTRUCTIONS] of var LOCATIONS: live_duration;\n'
        ret += 'array[INSTRUCTIONS] of int: output_register_count = [%s];\n' % ', '.join(str(reg_count(line_of_var(input_data, var)['type'])) for var in var_names)
        ret += 'var LOCATIONS: RET_loc;\n' 
        ret += 'constraint cumulative(output_locations, live_duration, output_register_count, %d);\n' % len(REGISTERS)
        ret += '\n'
        return ret

    def make_disjoint(input_data):
        var_names = get_var_names(input_data)
        ret = ''
        ret += 'constraint alldifferent(output_locations);\n'
        return ret

    def make_dependencies_use(input_data):
        ret = ''
        reverse_dependencies = {}
        for line in input_data['lines']:
            for arg in line['args']:
                if arg[:2] == '0x': continue
                if arg not in reverse_dependencies.keys(): reverse_dependencies[arg] = []
                reverse_dependencies[arg].append(line['out'])
        for var in reverse_dependencies.keys():
            ret += ('constraint output_locations[%s] + live_duration[%s] == max([%s]);\n'
                    % (var, var, ', '.join('output_locations[%s]' % dep for dep in reverse_dependencies[var])))
        ret += '\n'
        return ret

    def make_dependencies(input_data):
        var_names = get_var_names(input_data)
        ret = ''
        for line in input_data['lines']:
            for arg in line['args']:
                if arg[0] not in '0123456789':
                    ret += ('constraint output_locations[%s] + 1 <= output_locations[%s];\n'
                            % (arg, line['out']))
        ret += '\n'
        ret += 'constraint max([ output_locations[i] + 1 | i in INSTRUCTIONS ]) <= RET_loc;\n'
        ret += '\n'
        return ret

    def make_output(input_data):
        ret = 'solve minimize RET_loc;\n\n'
        ret += 'output [ "(" ++ show(INSTRUCTIONS_NAMES[i]) ++ ", " ++ show(output_locations[i]) ++ ", " ++ show(live_duration[i]) ++ ") ,\\n"\n'
        ret += '       | i in INSTRUCTIONS ];\n'
        ret += 'output [ "RET_loc: " ++ show(RET_loc) ];\n'
        return ret

    return '\n'.join([
        make_decls(data),
        make_disjoint(data),
        make_dependencies_use(data),
        make_dependencies(data),
        make_output(data)
        ])


RESULTS = [
"""("x20", "ADD_MUL", 9, 12, 4) ,
("x21", "ADD_MUL", 21, 12, 4) ,
("x22", "ADD_MUL", 10, 12, 4) ,
("x23", "ADD_MUL", 34, 4, 4) ,
("x24", "ADD_MUL", 47, 12, 4) ,
("x25", "ADD_MUL", 43, 12, 4) ,
("x26", "ADD_MUL", 59, 4, 4) ,
("x27", "ADD_MUL", 51, 12, 4) ,
("x28", "ADD_MUL", 63, 4, 4) ,
("x29", "ADD_MUL", 75, 12, 4) ,
("x30", "ADD_MUL", 71, 12, 4) ,
("x31", "ADD_MUL", 87, 4, 4) ,
("x32", "ADD_MUL", 79, 12, 4) ,
("x33", "ADD_MUL", 91, 4, 4) ,
("x34", "ADD_MUL", 82, 12, 4) ,
("x35", "ADD_MUL", 95, 4, 4) ,
("x36", "ADD_MUL", 58, 12, 4) ,
("x37", "ADD_MUL", 55, 12, 4) ,
("x38", "ADD_MUL", 70, 4, 4) ,
("x39", "ADD_MUL", 62, 12, 4) ,
("x40", "ADD_MUL", 74, 4, 4) ,
("x41", "ADD_MUL", 66, 12, 4) ,
("x42", "ADD_MUL", 78, 4, 4) ,
("x43", "ADD_MUL", 90, 12, 4) ,
("x44", "ADD_MUL", 102, 4, 4) ,
("x45", "ADD_MUL", 1, 12, 4) ,
("x46", "ADD_MUL", 2, 12, 4) ,
("x47", "ADD_MUL", 5, 12, 4) ,
("x48", "ADD_MUL", 6, 12, 4) ,
("x49", "ADD_MUL", 13, 12, 4) ,
("x50", "ADD_MUL", 25, 4, 4) ,
("x51", "ADD_MUL", 14, 12, 4) ,
("x52", "ADD_MUL", 29, 4, 4) ,
("x53", "ADD_MUL", 17, 12, 4) ,
("x54", "ADD_MUL", 33, 4, 4) ,
("x55", "ADD_MUL", 18, 12, 4) ,
("x56", "ADD_MUL", 37, 4, 4) ,
("x57", "ADD_MUL", 22, 12, 4) ,
("x58", "ADD_MUL", 38, 4, 4) ,
("x59", "ADD_MUL", 30, 12, 4) ,
("x60", "ADD_MUL", 42, 4, 4) ,
("x61", "ADD_MUL", 26, 12, 4) ,
("x62", "ADD_MUL", 46, 4, 4) ,
("x63", "ADD_MUL", 54, 12, 4) ,
("x64", "ADD_MUL", 67, 4, 4) ,
("x65", "ADD_MUL", 86, 12, 4) ,
("x66", "ADD_MUL", 98, 4, 4) ,
("x67", "ADD_MUL", 83, 12, 4) ,
("x68", "ADD_MUL", 99, 4, 4) ,
("x69", "LEA_BW", 41, 4, 4) ,
("x70", "LEA_BW", 44, 4, 4) ,
("x71", "ADD_MUL", 50, 4, 4) ,
("x72", "LEA_BW", 56, 4, 4) ,
RET_loc: 106"""
,
"""("x73", "LEA_BW", 2, 4, 4) ,
("x74", "ADD_MUL", 1, 4, 4) ,
("x75", "LEA_BW", 5, 4, 4) ,
("x76", "LEA_BW", 6, 4, 4) ,
("x77", "ADD_MUL", 9, 4, 4) ,
("x78", "LEA_BW", 13, 4, 4) ,
("x79", "LEA_BW", 14, 4, 4) ,
("x80", "ADD_MUL", 17, 4, 4) ,
("x81", "LEA_BW", 21, 4, 4) ,
("x82", "LEA_BW", 22, 4, 4) ,
("x83", "ADD_MUL", 25, 12, 4) ,
("x84", "ADD_MUL", 37, 4, 4) ,
("x85", "LEA_BW", 41, 4, 4) ,
("x86", "LEA_BW", 42, 4, 4) ,
("x87", "ADD_MUL", 45, 4, 4) ,
("x88", "LEA_BW", 49, 4, 4) ,
("x89", "LEA_BW", 50, 4, 4) ,
("x90", "ADD_MUL", 53, 4, 4) ,
RET_loc: 57"""
]

######################################################################
######################################################################
##                     Parsing minizinc output                      ##
######################################################################
######################################################################

def assemble_output_and_register_allocate(data_list, result_list):
    def parse_result(result):
        def parse_val(val):
            val = val.strip()
            if val[0] == '(' and val[-1] == ')':
                return tuple(map(parse_val, val[1:-1].split(',')))
            if val[0] in '"\'' and val[-1] in '"\'':
                return val[1:-1]
            if val.isdigit():
                return int(val)
            print('Unknown value: %s' % val)
            return val
        ret = {'schedule':list(map(str.strip, result.split(',\n')))}
        ret['RET_loc'] = [i[len('RET_loc: '):] for i in ret['schedule'] if i[:len('RET_loc: ')] == 'RET_loc: ']
        assert len(ret['RET_loc']) == 1
        ret['RET_loc'] = int(ret['RET_loc'][0])
        ret['schedule'] = tuple(parse_val(val) for val in ret['schedule'] if val[0] == '(' and val[-1] == ')')
        return ret
        
    def combine_lists(data_list, result_list):
        data = data_list[0]
        data['lines'] = list(data['lines'])
        for cur_data in data_list[1:]:
            data['lines'] += list(cur_data['lines'])
        data['lines'] = tuple(data['lines'])

        basepoint = 0
        results = []
        for result in map(parse_result, result_list):
            results += [(var, core_type, basepoint+loc, data_latency, core_latency)
                        for var, core_type, loc, data_latency, core_latency in result['schedule']]
            basepoint += result['RET_loc']
        return (data, results, basepoint)

    def sort_results(data, results):
        return sorted([(loc, line, var, core_type, data_latency, core_latency)
                       for (line, (var, core_type, loc, data_latency, core_latency))
                       in zip(data['lines'], results)])

    def get_live_ranges(data, results, RET_loc):
        var_names = get_var_names(data)
        input_var_names = get_input_var_names(data)
        output_var_names = get_output_var_names(data)
        ret = dict((var, {'start':0, 'accessed':[]}) for var in input_var_names)
        for (line, (var, core_type, loc, data_latency, core_latency)) in zip(data['lines'], results):
            assert var not in ret.keys()
            ret[var] = {'start':loc, 'accessed':[]}
            for arg in line['args']:
                if arg in var_names + input_var_names:
                    ret[arg]['accessed'].append(loc)
                else:
                    print(arg)
        for var in output_var_names:
            ret[var]['accessed'].append(RET_loc)
        for var in ret.keys():
            ret[var]['end'] = max(ret[var]['accessed'])
        return ret

    def remake_overlaps(live_ranges):
        live_ranges = dict(live_ranges)
        for var in live_ranges.keys():
            live_ranges[var] = dict(live_ranges[var])
            live_ranges[var]['overlaps'] = tuple(sorted(
                other_var
                for other_var in live_ranges.keys()
                if other_var != var and
                (live_ranges[other_var]['start'] <= live_ranges[var]['start'] <= live_ranges[other_var]['end']
                 or live_ranges[var]['start'] <= live_ranges[other_var]['start'] <= live_ranges[var]['end'])
                ))
        return live_ranges

##    def insert_possible_registers(live_ranges):
##        live_ranges = dict(live_ranges)
##        for var in live_ranges.keys():
##            live_ranges[var] = dict(live_ranges[var])
##            live_ranges[var]['possible registers'] = tuple(sorted(
##                other_var
##                for other_var in live_ranges.keys()
##                if other_var != var and
##                (live_ranges[other_var]['start'] <= live_ranges[var]['start'] <= live_ranges[other_var]['end']
##                 or live_ranges[var]['start'] <= live_ranges[other_var]['start'] <= live_ranges[var]['end'])
##                ))
##        return live_ranges

    def register_allocate(live_ranges):
        allocated = {}
        remaining_registers = list(REGISTERS)
        

    def format_results(sorted_results):
        return ['%s // %s, start: %.2f, end: %.2f' % (line['source'], core_type, loc / 4.0, (loc + data_latency) / 4.0)
                for loc, line, var, core_type, data_latency, core_latency in sorted_results]

    data, results, RET_loc = combine_lists(data_list, result_list)
##    live_ranges = remake_overlaps(get_live_ranges(data, results, RET_loc))
##    for i in sorted((len(rangev['overlaps']), var, rangev['accessed'], (rangev['start'], rangev['end']), rangev['overlaps']) for var, rangev in live_ranges.items()):
##        print(i)
    return '\n'.join(format_results(sort_results(data, results)))

data_list = parse_lines(get_lines('femulDisplay.log'))
for i, data in enumerate(data_list):
    with open('femulDisplayReg_%d.mzn' % i, 'w') as f:
        #f.write(make_assign_locations_to_instructions(data))
        #f.write(make_assign_instructions_to_locations(data))
        f.write(make_assign_locations_to_instructions_cumulatively(data))
            
#print(assemble_output_and_register_allocate(data_list, RESULTS))



import socket, sys
from cryptography.hazmat.primitives.asymmetric import x25519

b = bool(int(sys.argv[1]))

# x25519.X25519PrivateKey.generate()
sk = x25519.X25519PrivateKey.from_private_bytes(
        b'P\x9d\xfb\x9bl\x044\xb9lv\x81\x13\xf0I8%\x99CO5.<>C\xa2\xbbH\t\xe42\xd1N')
print (', '.join("0x%02x"%b for b in sk.public_key()._raw_public_bytes()))

s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
s.settimeout(1000.0)
s.sendto(b'\x01\x00' + b'A'*32, ('10.55.55.3', 1823))

try:
    data, remote = s.recvfrom(2048)
    print('pk:', data[2:][:32])
    assert len(data) >= 2+32
    pk = x25519.X25519PublicKey.from_public_bytes(data[2:][:32])
    ss = sk.exchange(pk)
    print('ss:', ss)
    s.sendto(b'\x02\x00'+(ss[:16], ss[16:])[b], ('10.55.55.3', 1823))
except socket.timeout:
    print('REQUEST TIMED OUT')



#!/usr/bin/env python
from __future__ import with_statement
from memoize import memoize
import codecs, re, sys, os
import subprocess

LAMBDA = u'\u03bb'

OP_NAMES = {'*':'MUL', '+':'ADD', '>>':'SHL', '<<':'SHR', '|':'OR', '&':'AND'}

REGISTERS = tuple(['RAX', 'RBX', 'RCX', 'RDX', 'RSI', 'RDI', 'RBP'] #, 'RSP'] # RSP is stack pointer?
                  + ['r%d' % i for i in range(8, 16)])
REGISTER_COLORS = ['color="black"', 'color="white",fillcolor="black"', 'color="maroon"', 'color="green"', 'fillcolor="olive"',
                   'color="navy"', 'color="purple"', 'fillcolor="teal"', 'fillcolor="silver"', 'fillcolor="gray"', 'fillcolor="red"',
                   'fillcolor="lime"', 'fillcolor="yellow"', 'fillcolor="blue"', 'fillcolor="fuschia"', 'fillcolor="aqua"']
REGISTER_COLORS = ['fillcolor="%s"' % c for c in 'aliceblue antiquewhite aquamarine azure beige bisque blue blueviolet brown cadetblue chartreuse cyan red gold deeppink darkorange'.split(' ')]
COLOR_FOR_REGISTER = dict(zip(REGISTERS, REGISTER_COLORS))

MAX_INSTRUCTION_WINDOW = 10000

def get_lines(filename):
    with codecs.open(filename, 'r', encoding='utf8') as f:
        lines = f.read().replace('\r\n', '\n')
    return [line.strip() for line in re.findall("%s '.*?[Rr]eturn [^\r\n]*" % LAMBDA, lines, flags=re.DOTALL)[0].split('\n')]

def strip_casts(text):
    return re.sub(r'\(u?int[0-9]*_t\)\s*\(?([^\)]*)\)?', r'\1', text)

def parse_lines(lines):
    lines = list(map(strip_casts, lines))
    assert lines[0][:len(LAMBDA + ' ')] == LAMBDA + ' '
    assert lines[0][-1] == ','
    ret = {}
    ret['vars'] = lines[0][len(LAMBDA + ' '):-1]
    assert lines[-1][-1] == ')'
    ret['return'] = lines[-1][:-1].replace('return ', '').replace('Return ', '')
    ret['lines'] = []
    for line in lines[1:-1]:
        datatype, varname, arg1, op, arg2 = re.findall('^(u?int[0-9]*_t) ([^ ]*) = ([^ ]*) ([^ ]*) ([^ ]*);$', line)[0]
        ret['lines'].append({'type':datatype, 'out':varname, 'op':op, 'args':(arg1, arg2), 'source':line})
    print('Compiling %d lines in groups of %d...' % (len(ret['lines']), min(MAX_INSTRUCTION_WINDOW, len(ret['lines']))))
    ret['lines'] = tuple(ret['lines'])
    split_ret = []
    for start in range(0, len(ret['lines']), MAX_INSTRUCTION_WINDOW):
        cur_ret = dict(ret)
        cur_ret['lines'] = ret['lines'][start:][:MAX_INSTRUCTION_WINDOW]
        split_ret.append(cur_ret)
    return tuple(split_ret)

def get_var_names(input_data):
    return tuple(line['out'] for line in input_data['lines'])

def get_input_var_names(input_data):
    return tuple(i for i in data['vars'].replace('%core', '').replace(',', ' ').replace('(', ' ').replace(')', ' ').replace("'", ' ').split(' ')
                 if i != '')

def get_output_var_names(input_data):
    return tuple(i for i in data['return'].replace(',', ' ').replace('(', ' ').replace(')', ' ').split(' ')
                 if i != '')

def line_of_var(input_data, var):
    retv = [line for line in input_data['lines'] if line['out'] == var]
    if len(retv) > 0: return retv[0]
    return {'out': var, 'args':tuple(), 'op': 'INPUT', 'type':'uint64_t'}

def create_set(name, items):
    ret = ''
    ret += 'set of int: %s = 1..%d;\n' % (name, len(items))
    for i, item in enumerate(items):
        ret += '%s: %s = %d; ' % (name, item, i+1)
    ret += 'array[%s] of string: %s_NAMES = ["' % (name, name)
    ret += '", "'.join(items) + '"];\n'
    ret += '\n'
    return ret

def make_data_dependencies(input_data):
    input_var_names = get_input_var_names(input_data)
    dependencies = dict((var, tuple()) for var in input_var_names)
    for line in input_data['lines']:
        dependencies[line['out']] = tuple(arg for arg in line['args']
                                          if arg[0] not in '0123456789')
    return dependencies
def make_reverse_data_dependencies(dependencies):
    ret = {}
    for k, vs in dependencies.items():
        for v in vs:
            if v not in ret.keys(): ret[v] = []
            ret[v].append(k)
    for k in ret.keys():
        ret[k] = tuple(ret[k])
    return ret

def make_reverse_data_dependencies(dependencies):
    reverse_dependencies = dict((k, []) for k in dependencies.keys())
    for k, v in dependencies.items():
        for arg in v:
            reverse_dependencies[arg].append(k)
    return reverse_dependencies

def backpropogate_one(input_data, dependencies, registers):
    progressed = False
    for var in registers.keys():
        if len(dependencies[var]) == 1:
            dep = dependencies[var][0]
            if dep not in registers.keys():
                line = line_of_var(input_data, dep)
                if line['type'] == 'uint64_t' and ':' not in registers[var]:
                    registers[dep] = registers[var]
                    progressed = True
            elif registers[dep] != registers[var] and registers[var] not in registers[dep].split(':') and len(dependencies[dep]) == 2:
                for dep2, other_dep2 in (tuple(dependencies[dep]), tuple(reversed(dependencies[dep]))):
                    if dep2 in registers.keys(): continue
                    if other_dep2 in registers.keys() and registers[other_dep2] == registers[var]: continue
                    line = line_of_var(input_data, dep2)
                    other_line = line_of_var(input_data, other_dep2)
                    if line['type'] == 'uint64_t' and ':' not in registers[var] and (other_dep2 in registers.keys() or other_line['type'] != 'uint64_t'):
                        registers[dep2] = registers[var]
                        progressed = True
        elif len(dependencies[var]) == 2:
            for dep, other_dep in (tuple(dependencies[var]), tuple(reversed(dependencies[var]))):
                if dep in registers.keys(): continue
                if other_dep in registers.keys() and registers[other_dep] == registers[var]: continue
                line = line_of_var(input_data, dep)
                if line['type'] == 'uint64_t' and ':' not in registers[var] and len(dependencies[dep]) == 1 and dependencies[dep][0] in registers.keys():
                    registers[dep] = registers[var]
                    progressed = True
    return progressed, registers
def backpropogate_one_arbitrary(input_data, dependencies, registers):
    progressed = False
    for var in registers.keys():
        if len(dependencies[var]) == 2:
            for dep, other_dep in (tuple(dependencies[var]), tuple(reversed(dependencies[var]))):
                if dep in registers.keys(): continue
                if other_dep in registers.keys() and (registers[other_dep] == registers[var] or registers[var] in registers[other_dep].split(':')): continue
                line = line_of_var(input_data, dep)
                if line['type'] == 'uint64_t' and ':' not in registers[var]:
                    registers[dep] = registers[var]
                    progressed = True
                elif line['type'] == 'uint128_t' and ':' in registers[var] and other_dep not in registers.keys():
                    registers[dep] = registers[var]
                    progressed = True
    return progressed, registers
def backpropogate_128(input_data, dependencies, reverse_dependencies, registers):
    progressed = False
    for var in registers.keys():
        if len(dependencies[var]) == 1 and len(reverse_dependencies[dependencies[var][0]]) == 2:
            var = dependencies[var][0]
            if var in registers.keys(): continue
            for dep, other_dep in (tuple(reverse_dependencies[var]), tuple(reversed(reverse_dependencies[var]))):
                if dep not in registers.keys() or other_dep not in registers.keys(): continue
                line = line_of_var(input_data, dep)
                other_line = line_of_var(input_data, other_dep)
                var_line = line_of_var(input_data, var)
                if var_line['type'] == 'uint128_t' and line['type'] == 'uint64_t' and other_line['type'] == 'uint64_t' and line['op'] == '>>' and other_line['op'] == '&':
                    registers[var] = registers[dep] + ':' + registers[other_dep]
                    progressed = True
    return progressed, registers
def backpropogate_one128(input_data, dependencies, registers):
    progressed = False
    for var in registers.keys():
        var_line = line_of_var(input_data, var)
        if var_line['type'] != 'uint128_t': continue
        if len(dependencies[var]) == 1:
            dep = dependencies[var][0]
            if dep not in registers.keys():
                line = line_of_var(input_data, dep)
                if line['type'] == 'uint128_t' and ':' in registers[var]:
                    registers[dep] = registers[var]
                    progressed = True
        elif len(dependencies[var]) == 2:
            for dep, other_dep in (tuple(dependencies[var]), tuple(reversed(dependencies[var]))):
                if dep in registers.keys(): continue
                if other_dep in registers.keys() and registers[other_dep] == registers[var]: continue
                line = line_of_var(input_data, dep)
                other_line = line_of_var(input_data, other_dep)
                if line['type'] == 'uint128_t' and other_line['type'] == 'uint64_t' and other_dep not in registers.keys() and ':' in registers[var]:
                    registers[dep] = registers[var]
                    progressed = True
                elif line['type'] == 'uint64_t' and other_line['type'] == 'uint64_t' and other_dep not in registers.keys() and ':' in registers[var] and var_line['op'] == '*':
                    registers[dep], registers[other_dep] = registers[var].split(':')
                    progressed = True
    return progressed, registers
def all_reverse_dependencies(reverse_dependencies, var_set):
    ret = set(var_set)
    for v in var_set:
        ret = ret.union(set(reverse_dependencies[v]))
    if len(ret) == len(set(var_set)): return ret
    return all_reverse_dependencies(reverse_dependencies, ret)
def unassigned_reverse_dependencies(reverse_dependencies, registers, var_set):
    return set(v for v in all_reverse_dependencies(reverse_dependencies, var_set) if v not in registers.keys())
def assign_one_new_reg(input_data, dependencies, reverse_dependencies, registers, new_reg):
    for var in registers.keys():
        var_line = line_of_var(input_data, var)
        for dep in dependencies[var]:
            if len(unassigned_reverse_dependencies(reverse_dependencies, registers, [dep])) == 1 and line_of_var(input_data, dep)['type'] == 'uint64_t':
                registers[dep] = new_reg
                return True, registers
    return False, registers

def assign_registers(input_data, dependencies):
    reverse_dependencies = make_reverse_data_dependencies(dependencies)
    registers = {}
    registers_available = list(REGISTERS)
    out_vars = get_output_var_names(input_data)
    for var in out_vars:
        registers[var] = registers_available.pop()
    progressed = True
    while progressed:
        progressed, registers = backpropogate_one(input_data, dependencies, registers)
    progressed1, progressed2 = True, True
    while progressed1 or progressed2:
        progressed1, registers = backpropogate_one(input_data, dependencies, registers)
        progressed2, registers = backpropogate_one_arbitrary(input_data, dependencies, registers)
    progressed5 = True
    max_count = 7
    c = 0
    while c < max_count and progressed5:
        progressed1, progressed2, did_progress = True, True, True
        while progressed1 or progressed2 or did_progress:
            progressed3, progressed4 = True, True
            while progressed3 or progressed4:
                progressed3, registers = backpropogate_128(input_data, dependencies, reverse_dependencies, registers)
                progressed4, registers = backpropogate_one128(input_data, dependencies, registers)
                did_progress = progressed3 or progressed4
            progressed1, registers = backpropogate_one(input_data, dependencies, registers)
            progressed2, registers = backpropogate_one_arbitrary(input_data, dependencies, registers)
        c += 1
        if c < max_count:
            reg, registers_available = registers_available[-1], registers_available[:-1]
            progressed5, registers = assign_one_new_reg(input_data, dependencies, reverse_dependencies, registers, reg)
    return registers

def print_dependencies(input_data, dependencies):
    in_vars = get_input_var_names(input_data)
    out_vars = get_output_var_names(input_data)
    registers = assign_registers(input_data, dependencies)
    body = (
        ''.join('    %s [label="%s (%s)",%s];\n' % (var, var, reg, COLOR_FOR_REGISTER[reg.split(':')[0]]) for var, reg in registers.items()) +
        ''.join('    in -> %s ;\n' % var for var in in_vars) +
        ''.join('    %s -> out ;\n' % var for var in out_vars) +
        ''.join(''.join('    %s -> %s ;\n' % (out_var, in_var) for out_var in sorted(dependencies[in_var]))
                for in_var in sorted(dependencies.keys()))
            )
    return ('digraph G {\n' + body + '}\n')
def adjust_bits(input_data, graph):
    for line in input_data['lines']:
        if line['type'] == 'uint128_t':
            graph = graph.replace(line['out'], line['out'] + '_128')
    return graph


data_list = parse_lines(get_lines('femulDisplay.log'))
for i, data in enumerate(data_list):
    deps = adjust_bits(data, print_dependencies(data, make_data_dependencies(data)))
    with codecs.open('femulData%d.dot' % i, 'w', encoding='utf8') as f:
        f.write(deps)
    for fmt in ('png', 'svg'):
        subprocess.call(['dot', '-T%s' % fmt, 'femulData%d.dot' % i, '-o', 'femulData%d.%s' % (i, fmt)])



#!/usr/bin/env python
from __future__ import with_statement
from memoize import memoize
import codecs, re, sys, os
import subprocess

LAMBDA = u'\u03bb'

OP_NAMES = {'*':'MUL', '+':'ADD', '>>':'SHL', '<<':'SHR', '|':'OR', '&':'AND'}

REGISTERS = tuple(['RAX', 'RBX', 'RCX', 'RDX', 'RSI', 'RDI', 'RBP', 'RSP']
                  + ['r%d' % i for i in range(8, 16)])

MAX_INSTRUCTION_WINDOW = 10000

CORE_DATA = (('ADD_MUL', 2), ('MUL_CORE', 1), ('LEA_BW', 2))
CORES = tuple(name for name, count in CORE_DATA)
CORE_COUNT = dict(CORE_DATA)

BITWISE_CORES = tuple({
    'core' : { 'name' : core_name , 'latency' : 1 },
    'latency' : 1
    } for core_name in ('LEA_BW',))

MODEL = {
    '*': tuple({
        'core' : { 'name' : core_name , 'latency' : 1 },
        'latency' : 3
        }
          for core_name in ('ADD_MUL', 'MUL_CORE')),
    '+': tuple({
        'core' : { 'name' : core_name , 'latency' : 1 },
        'latency' : 1
        }
          for core_name in ('ADD_MUL', 'LEA_BW')),
    '>>': BITWISE_CORES,
    '<<': BITWISE_CORES,
    '|': BITWISE_CORES,
    '&': BITWISE_CORES,
    'LOAD': tuple({
        'core' : { 'name' : core_name , 'latency' : 1 },
        'latency' : 1
        } for core_name in REGISTERS),
    'STORE': tuple({
        'core' : { 'name' : core_name , 'latency' : 1 },
        'latency' : 1
        } for core_name in REGISTERS)
    }

def get_lines(filename):
    with codecs.open(filename, 'r', encoding='utf8') as f:
        lines = f.read().replace('\r\n', '\n')
    return [line.strip() for line in re.findall("%s '.*?[Rr]eturn [^\r\n]*" % LAMBDA, lines, flags=re.DOTALL)[0].split('\n')]

def strip_casts(text):
    return re.sub(r'\(u?int[0-9]*_t\)\s*\(?([^\)]*)\)?', r'\1', text)

def parse_lines(lines):
    lines = list(map(strip_casts, lines))
    assert lines[0][:len(LAMBDA + ' ')] == LAMBDA + ' '
    assert lines[0][-1] == ','
    ret = {}
    ret['vars'] = lines[0][len(LAMBDA + ' '):-1]
    assert lines[-1][-1] == ')'
    ret['return'] = lines[-1][:-1].replace('return ', '').replace('Return ', '')
    ret['lines'] = []
    for line in lines[1:-1]:
        datatype, varname, arg1, op, arg2 = re.findall('^(u?int[0-9]*_t) ([^ ]*) = ([^ ]*) ([^ ]*) ([^ ]*);$', line)[0]
        ret['lines'].append({'type':datatype, 'out':varname, 'op':op, 'args':(arg1, arg2), 'source':line})
    print('Compiling %d lines in groups of %d...' % (len(ret['lines']), min(MAX_INSTRUCTION_WINDOW, len(ret['lines']))))
    ret['lines'] = tuple(ret['lines'])
    split_ret = []
    for start in range(0, len(ret['lines']), MAX_INSTRUCTION_WINDOW):
        cur_ret = dict(ret)
        cur_ret['lines'] = ret['lines'][start:][:MAX_INSTRUCTION_WINDOW]
        split_ret.append(cur_ret)
    return tuple(split_ret)

def get_var_names(input_data):
    return tuple(line['out'] for line in input_data['lines'])

def get_input_var_names(input_data):
    return tuple(i for i in data['vars'].replace('%core', '').replace(',', ' ').replace('(', ' ').replace(')', ' ').replace("'", ' ').split(' ')
                 if i != '')

def get_output_var_names(input_data):
    return tuple(i for i in data['return'].replace(',', ' ').replace('(', ' ').replace(')', ' ').split(' ')
                 if i != '')

def create_set(name, items):
    ret = ''
    ret += 'set of int: %s = 1..%d;\n' % (name, len(items))
    for i, item in enumerate(items):
        ret += '%s: %s = %d; ' % (name, item, i+1)
    ret += 'array[%s] of string: %s_NAMES = ["' % (name, name)
    ret += '", "'.join(items) + '"];\n'
    ret += '\n'
    return ret

def make_data_dependencies(input_data):
    input_var_names = get_input_var_names(input_data)
    dependencies = dict((var, tuple()) for var in input_var_names)
    for line in input_data['lines']:
        dependencies[line['out']] = tuple(arg for arg in line['args']
                                          if arg[0] not in '0123456789')
    return dependencies

def make_reverse_data_dependencies(dependencies):
    reverse_dependencies = {}
    for k, v in dependencies.items():
        for arg in v:
            if arg not in reverse_dependencies.keys(): reverse_dependencies[arg] = []
            reverse_dependencies[arg].append(k)
    return reverse_dependencies

def print_dependencies(input_data, dependencies):
    in_vars = get_input_var_names(input_data)
    out_vars = get_output_var_names(input_data)
    return ('digraph G {\n' +
            ''.join('    in -> %s ;\n' % var for var in in_vars) +
            ''.join('    %s -> out ;\n' % var for var in out_vars) +
            ''.join(''.join('    %s -> %s ;\n' % (out_var, in_var) for out_var in sorted(dependencies[in_var]))
                    for in_var in sorted(dependencies.keys())) +
            '}\n')
def adjust_bits(input_data, graph):
    for line in input_data['lines']:
        if line['type'] == 'uint128_t':
            graph = graph.replace(line['out'], line['out'] + '_128')
    return graph
    

data_list = parse_lines(get_lines('femulDisplay.log'))
for i, data in enumerate(data_list):
    deps = adjust_bits(data, print_dependencies(data, make_data_dependencies(data)))
    with codecs.open('femulData%d.dot' % i, 'w', encoding='utf8') as f:
        f.write(deps)
    for fmt in ('png', 'svg'):
        subprocess.call(['dot', '-T%s' % fmt, 'femulData%d.dot' % i, '-o', 'femulData%d.%s' % (i, fmt)])
